import numpy as np


def exploit(x, model, params):
    """
    Purely exploiting acquisition function

    Doesn't take any parameters (apart from the model).
    """
    m, s, dmdx, dsdx = model.predictive_m_s_grad(x)
    f_acqu = m
    df_acqu = dmdx
    scipygradient = np.asmatrix(df_acqu).transpose()
    return f_acqu, scipygradient



import sys

from boss import __version__
from boss.bo.bo_main import BOMain
from boss.settings import Settings
from boss.io.main_output import MainOutput
from boss.mep.mepmain import MEPMain
from boss.pp.pp_main import PPMain
from boss.bo.rstmanager import RstManager
from boss.utils.timer import Timer
from boss.io.parse import parse_input_file


def main(args=None):
    """The main routine."""
    # start timers
    local_timer = Timer()

    if args is None:
        args = sys.argv[1:]

    if not args_ok(args):  # Exit immediately if one or more args are invalid.
        print(
            "BOSS version "
            + str(__version__)
            + "\n"
            + "Usage:\n"
            + "   boss op <inputfile or rst-file>\n"
            + "   boss o <inputfile or rst-file>\n"
            + "   boss p <rst-file> <out-file>\n"
            + "   boss m <rst-file> <minima-file>\n"
            + "See the documentation for further instructions."
        )
        return

    if not files_ok(args[1:]):  # Exit immediately if input file doesn't open.
        return

    input_data = parse_input_file(args[1])
    settings = Settings(input_data["keywords"])

    # Don't overwrite an optimization run's outfile.
    if len(args) == 3:
        ipt_outfile = args[2]
        if "m" in args[0]:
            settings["outfile"] = settings["outfile"][:-4] + "_mep.out"
        else:
            settings["outfile"] = settings["outfile"][:-4] + "_pp.out"

    main_output = None

    # 1. Run bayesian optimization. Note: if we run an optimization we let BOMain
    # handle the MainOutput
    if "o" in args[0] and (settings["initpts"] + settings["iterpts"]) > 0:
        local_timer.startLap()

        rst_data = input_data.get("rst_data", None)
        bo = BOMain.from_settings(settings, rst_data)
        main_output = bo.main_output
        bo.run()

        main_output.progress_msg(
            "| Bayesian optimization completed, "
            + "time [s] %s" % (local_timer.str_lapTime()),
            1,
            True,
            True,
        )

    # If no optmization was run, we need to initialize the MainOutput and
    # start a new file manually
    if not main_output:
        main_output = MainOutput(settings)
        main_output.new_file()

    # 2. Run post-processing.
    if "p" in args[0] and len(settings["pp_iters"]) > 0:
        local_timer.startLap()
        main_output.progress_msg("Starting post-processing...", 1, True)
        main_output.section_header("POST-PROCESSING")

        ipt_rstfile = settings["rstfile"] if "o" in args[0] else args[1]
        if len(args) != 3:
            ipt_outfile = settings["outfile"]

        settings["rstfile"] = ipt_rstfile
        settings["outfile"] = ipt_outfile

        pp_main = PPMain.from_settings(settings, main_output)
        pp_main.run()

        main_output.progress_msg(
            "Post-processing completed, " + "time [s] %s" % (local_timer.str_lapTime()),
            1,
        )

    # 3. Find minimum energy paths.
    if "m" == args[0]:
        local_timer.startLap()
        main_output.progress_msg("Finding minimum energy paths...", 1, True)
        main_output.section_header("MINIMUM ENERGY PATHS")

        MEPMain(settings, args[1], args[2], main_output)
        main_output.progress_msg(
            "Finding minimum energy paths completed, "
            + "time [s] %s" % (local_timer.str_lapTime()),
            1,
        )

    main_output.footer(local_timer.str_totalTime())


def args_ok(args):
    """
    Checks that the user has called BOSS properly by examining the arguments
    given, number of files and filename extensions. BOSS should be called with
    one of the following:
        boss o options/.rst-file
        boss op options/.rst-file
        boss p .rst-file .out-file
        boss m .rst-file local_minima.dat
    """
    # TODO prevent calling boss pm
    some_args = len(args) > 0
    if some_args:
        optim_ok = "o" in args[0] and len(args) == 2
        justpp_arg_ok = "o" not in args[0] and "p" in args[0] and len(args) == 3
        mep_arg_ok = "o" not in args[0] and "m" in args[0] and len(args) == 3
        rst_incl = len(args) >= 2 and ".rst" in args[1]
        out_incl = len(args) == 3 and ".out" in args[2]
        dat_incl = len(args) == 3 and ".dat" in args[2]
        justpp_ok = justpp_arg_ok and rst_incl and out_incl
        mep_ok = mep_arg_ok and rst_incl and dat_incl
    return some_args and (optim_ok or justpp_ok or mep_ok)


def files_ok(filenames):
    """
    Checks that the given files exist and can be opened.
    """
    for fname in filenames:
        try:
            f = open(fname, "r")
            f.close()
        except FileNotFoundError:
            print("Could not find file '" + fname + "'")
            return False
    return True


# Start BOSS
if __name__ == "__main__":
    main()



import numpy as np


class Space:
    def __init__(self, bounds, pbc):
        # bounds = [[lowest_crd1, lowest_crd2,...], [highest1, highest2, ...]]
        # pbc = [is_dim1_periodic, is_dim2_periodic, ...]
        self.dim = len(bounds[0])
        self.bounds = bounds
        self.bounds = self.bounds.astype(np.float)
        self.pbcdims = np.array(range(self.dim))[np.array(pbc)]
        self.nonpbcdims = np.array(range(self.dim))[np.logical_not(pbc)]

    def vec(self, to, frompt):
        # calculate vector1 (to) - vector2 (frompt) accounting for pbc
        result = to - frompt
        for j in self.pbcdims:
            a = -(self.bounds[1][j] - self.bounds[0][j]) / 2
            l = self.bounds[1][j] - self.bounds[0][j]
            result[j] = (result[j] - a) % l + a

        return result

    def crds(self, points):
        # replace any points outside of range of a nonpbc dim with nan
        # project any points outside of range of a pbc dim to within bounds
        for j in self.nonpbcdims:
            points[points[:, j] < self.bounds[0][j], :] = np.nan
            points[points[:, j] > self.bounds[1][j], :] = np.nan

        for j in self.pbcdims:
            a = self.bounds[0][j]
            l = self.bounds[1][j] - a
            points[:, j] = (points[:, j] - a) % l + a

        return points

    def trace(self, start, target, bo, eps, max_e):
        direction = self.vec(target, start)
        length = np.linalg.norm(direction)
        if length < eps:
            return target
        for i in range(1, int(length / eps)):
            test_point = start + i * eps * direction / np.linalg.norm(direction)
            test_point = self.crds(np.array([test_point]))[0]
            energy = bo.get_mu(np.vstack(np.atleast_2d(test_point)))[0, 0]
            if energy > max_e:
                l = np.linalg.norm(direction)
                prev_point = start + (i - 1) * eps * direction / l
                prev_point = self.crds(np.array([prev_point]))[0]
                return prev_point
        return target



import copy
import importlib
import os
import sys
from inspect import getfile
from pathlib import Path

import numpy as np

categories = {}
categories[(bool, 0)] = {
    "ygrads": False,
    "pp_models": False,
    "pp_acq_funcs": False,
    "pp_truef_at_xhats": False,
    "initupdate": True,
}
categories[(int, 0)] = {
    "initpts": 5,
    "verbosity": 1,
    "iterpts": None,
    "updatefreq": 1,
    "cores": 1,
    "updaterestarts": 2,
    "updateoffset": 0,
    "mep_precision": 25,
    "mep_rrtsteps": 10000,
    "mep_nebsteps": 20,
    "pp_truef_npts": None,
    "hmciters": 0,
    "minzacc": 15,
    "minfreq": 1,
    "userfn_arrdim": 2,
    "seed": None,
}
categories[(int, 1)] = {
    "pp_iters": None,
    "pp_model_slice": None,
}
categories[(float, 0)] = {
    "noise": 1e-12,
    "acqtol": 0.001,
    "min_dist_acqs": None,
    "mep_maxe": None,
    "pp_local_minima": None,
}
categories[(float, 1)] = {
    "yrange": [-10.0, 10.0],
    "gm_tol": None,
    "periods": None,
    "acqfnpars": np.array([]),
    "pp_var_defaults": None,
    "thetainit": None,
}
categories[(float, 2)] = {"bounds": None, "thetapriorpar": None, "thetabounds": None}
categories[(str, 0)] = {
    "outfile": "boss.out",
    "ipfile": "boss.in",
    "rstfile": "boss.rst",
    "acqfn_name": "elcb",
    "inittype": "sobol",
    "thetaprior": "gamma",
    "userfn": None,
}
categories[(str, 1)] = {
    "kernel": ["rbf"],
}


def get_copied_categories():
    """Returns deep copies of all categories.

    Useful when setting default values using the category 
    dicts and we want to avoid changing the default value
    contained in the category dict itself.
    """
    return [copy.deepcopy(cat) for cat in categories.values()]


def find_category(key):
    """Finds the category dict that contains a given key. """
    for cat, cat_dict in categories.items():
        if key in cat_dict:
            return cat
    raise KeyError(f"No category correspondig to keyword: {key}.")


def _eval_bool(x):
    """Converts string input to booleans. 

    Boolean values can be specified in BOSS input files as 
    0 / 1, [y]es / [n]o, [t]rue / [f]alse where all the words are 
    case-insensitive. This function handles conversion from these strings
    to proper Python booleans.
    """
    truthy = bool(x == "1" or x.lower()[0] == "y" or x.lower()[0] == "t")
    falsey = bool(x == "0" or x.lower()[0] == "n" or x.lower()[0] == "f")
    return truthy and not falsey


def destringify(val_str, category):
    """Converts a string to an appropriate Python object.

    When a boss input file is parsed, each string containing the value of a keyword
    is passed to this function. Strings are evaluated according to the BOSS input
    ruleset (see BOSS documentation).

    Parameters
    ----------
    val_str : str
        An input string to be evaluated.
    category : Tuple[type, int]
        The target type and dimensionality of the string evaluation. For instance,
        a string 'True' with category (bool, 1) will be evaluated to [True].

    Returns
    ----------
    Any
        The result of the string evaluation.
    """
    cat_type, cat_dim = category[0], category[1]
    val_str = val_str.strip()
    if val_str.lower() == "none":
        val = None
    elif cat_dim == 0:
        if cat_type is bool:
            val = _eval_bool(val_str)
        else:
            val = cat_type(val_str)
    elif cat_dim == 1:
        val_split = val_str.split()
        if cat_type == bool:
            val = np.asarray([_eval_bool(x) for x in val_split])
        else:
            val = [cat_type(x) for x in val_split]
            if cat_type in [int, float]:
                val = np.asarray(val)
    elif cat_dim == 2:
        rows = val_str.split(";")
        val = np.asarray([np.fromstring(row, sep=" ") for row in rows])
    else:
        raise ValueError(f"Cannot convert '{val_str}' to {category}")
    return val


def stringify(val):
    """Convert a Python type to a BOSS-style string.

    Parameters
    ----------
    val : str
        Python object to stringify
    """
    val_ndim = np.ndim(val)
    if val_ndim == 0:
        if isinstance(val, bool):
            val_str = str(int(val))
        else:
            val_str = str(val)
    if val_ndim == 1:
        if len(val) == 0:
            val_str = None
        else:
            val_str = " ".join([str(x) for x in val])
    elif val_ndim >= 2:
        val = np.array(val)
        val_str = str(val)
        val_str = val_str.replace("\n", ";")
        val_str = val_str.replace("[", "")
        val_str = val_str.replace("]", "")
    return val_str


def func_from_keyword(func_str):
    """
    """
    fsplit = func_str.split()
    if len(fsplit) == 2:
        func_name = fsplit[1]
    else:
        func_name = "f"
    func_path = Path(fsplit[0])
    sys.path.append(str(func_path.parent))
    func = getattr(importlib.import_module(func_path.stem), func_name)
    return func


def func_to_keyword(func):
    """
    The user function is either an ordinary function or a user defined,
    callable object. In the second case we must pass the type and not the
    object instance to inspect.getfile.
    """
    if type(func).__name__ == "function":
        func_path = getfile(func)
        func_name = func.__name__
    else:
        func_path = getfile(type(func))
        func_name = type(func).__name__
    path_str = f"{func_path} {func_name}"
    return path_str



import os.path
import numpy as np
import boss.io.ioutils as io
import boss.io.dump as dump


def split_rst_data(rst_data, dim):
    """Split the restart data into acquisitions and model parameters.

    Parameters
    ----------
    rst_data: np.ndarray
        The rst_data containing rows of (X, y, theta),
        e.g., as returned by parse_input_data.
    dim: int
        The dimension of the X-data.

    Returns
    -------
    acqs: np.ndarray
        Array containing the interation number (starting from 1) and the correspoding
        acquisitions found in the restart data.
    mod_par: np.ndarray
        Array containing the interation number (starting from 1) and the correspoding
        model parameters found in the restart data.
    """
    totalpts = rst_data.shape[0]
    # The number of input pts w/o mod param values is the number of rows with at most
    # dim + 1 entries that are not np.nan.
    inputpts = np.sum(np.sum(~np.isnan(rst_data), axis=1) <= dim + 1)
    acqs = np.c_[np.arange(1, totalpts + 1), rst_data[:, : dim + 1]]
    if totalpts > inputpts:
        mod_par = np.c_[
            np.arange(inputpts + 1, totalpts + 1), rst_data[inputpts:, dim + 1 :],
        ]
    else:
        mod_par = np.array([])
    return acqs, mod_par


class RstManager:
    """
    A class that handles restart-files (rst-files). These files can be used to
        1) introduce acquisition data from another source as initial values,
        2) continue a run that has been interrupted for some reason,
        3) continue a finished run by acquiring more points or
        4) keep the acquisitions but change the model or settings for a rerun.
    """

    def __init__(self, settings, rst_data=None):
        """
        Initializes the class with an array (settings.rstvals) containing the data
        read from an rst-file.
        """
        if rst_data is None:
            rst_data = np.array([])
        self.data = rst_data
        self.settings = settings
        self.x_dim = settings.dim
        self.use_gradients = settings["ygrads"]
        self.rstfile = settings["rstfile"]

    def new_file(self):
        if os.path.isfile(self.rstfile):
            print("warning: overwriting file '" + self.rstfile + "'")
        dump.dump_input_file(self.rstfile, self.settings, results_header=True)

    def get_x(self, i):
        """
        Returns the i:th acquisition location from the rst-data or None if
        it can't be found.
        """
        if (
            self.data.shape[0] > i
            and self.data.shape[1] >= self.x_dim
            and np.sum(np.isnan(self.data[i, 0 : self.x_dim])) == 0
        ):
            x_new = self.data[i, 0 : self.x_dim]
            return x_new
        else:
            return None

    def get_y(self, i):
        """
        Returns the i:th acquisition evaluation (energy and gradient) from the
        rst-data or None if it can't be found.
        """
        len_x = self.x_dim  # for convenience
        len_y = 1
        if self.use_gradients:
            len_y += len_x
        y_new = None
        yd_new = None
        # Actual functionality:
        if (
            self.data.shape[0] > i
            and self.data.shape[1] >= len_x + len_y
            and sum(np.isnan(self.data[i, len_x : (len_x + len_y)])) == 0
        ):
            y_new = self.data[i, len_x]
            if self.use_gradients:
                yd_new = self.data[i, (len_x + 1) : (len_x + len_y)]
        if not self.use_gradients:
            return y_new
        else:
            return (y_new, yd_new)

    def get_theta(self, i, n):
        """
        Returns the model paramters at iteration i from the rst-data or None
        if they can't be found.
        """
        len_x = self.x_dim  # for convenience
        len_y = 1
        if self.use_gradients:
            len_y += len_x
        a = len_x + len_y

        # Actual functionality:
        if (
            self.data.shape[0] > i
            and self.data.shape[1] == a + n
            and np.sum(np.isnan(self.data[i, a : (a + n)]) == 0)
        ):
            return self.data[i, a : (a + n)]
        else:
            return None

    def new_data(self, x, y):
        """
        Outputs a new data point (x,y) to rst file.
        """
        rst = open(self.rstfile, "a")
        rst.write("\n" + io.data_line(x, y, fstr="%23.15E")[:-1])
        rst.close()

    def new_model_params(self, mod_param):
        """
        Outputs a new set of model parameters to rst file.
        Format is variance, lengthscales
        """
        sigma = mod_param[0]
        lss = mod_param[1:]
        rst = open(self.rstfile, "a")
        rst.write("     " + io.data_line(sigma, lss, fstr="%23.15E")[:-1])
        rst.close()



import os
import shutil
import warnings

import numpy as np
from scipy.spatial.distance import euclidean

import boss.io.dump as dump
import boss.io.ioutils as ioutils
import boss.io.parse as parse
import boss.pp.plot as plot
from boss.bo.bo_main import BOMain
from boss.bo.model import Model
from boss.io.main_output import MainOutput
from boss.settings import Settings
from boss.utils.minimization import Minimization
from boss.bo.rstmanager import split_rst_data

warnings.filterwarnings("ignore")  # ignore warnings


def recreate_bo(settings, acqs, par):
    """Loads given data and parameters to a given model object. """
    dim = settings.dim
    X = acqs[:, :dim]
    Y = acqs[:, dim : dim + 1]
    bo = BOMain.from_settings(settings)
    bo.add_xy_list(X, Y)
    bo.model.set_unfixed_params(par)
    return bo


def _find_index(array, npts):
    if array.size == 0:
        return None
    else:
        ind = np.where(array[:, 0] == npts)[0]
        if ind.size == 0:
            return None
        else:
            return ind[-1]


class PPMain:
    """Performs the automated post-processing of a BOSS run. """

    def __init__(
        self,
        bo_results,
        pp_iters=None,
        pp_acq_funcs=None,
        pp_models=None,
        pp_model_slice=None,
        pp_var_defaults=None,
        pp_truef_at_xhats=None,
        pp_truef_npts=None,
        pp_local_minima=None,
    ):
        pp_settings = {
            "pp_iters": pp_iters,
            "pp_acq_funcs": pp_acq_funcs,
            "pp_models": pp_models,
            "pp_model_slice": pp_model_slice,
            "pp_var_defaults": pp_var_defaults,
            "pp_truef_at_xhats": pp_truef_at_xhats,
            "pp_truef_npts": pp_truef_npts,
            "pp_local_minima": pp_local_minima,
        }
        bo_results.settings.update(
            {k: v for k, v in pp_settings.items() if v is not None}
        )
        bo_results.settings.correct()

        self.bo_results = bo_results
        self.settings = bo_results.settings
        self.rstfile = self.settings["rstfile"]
        self.outfile = self.settings["outfile"]
        self.main_output = None
        self._setup()

    @classmethod
    def from_file(cls, rstfile, outfile, main_output=None):
        self = cls.__new__(cls)
        self.bo_results = None
        self.settings = Settings.from_file(rstfile)
        self.rstfile = rstfile
        self.outfile = outfile
        self.main_output = main_output
        cls._setup(self)
        return self

    @classmethod
    def from_settings(cls, settings, main_output=None):
        self = cls.__new__(cls)
        self.bo_results = None
        self.settings = settings
        self.rstfile = settings["rstfile"]
        self.outfile = settings["outfile"]
        self.main_output = main_output
        cls._setup(self)
        return self

    def _setup(self):
        """Common setup for all factory methods. """

        # User-specified model slices are interpreted as being indexed
        # from 1, so we convert to 0-based indexing here.
        self.settings["pp_model_slice"][:2] -= 1

        if not getattr(self, "main_output", None):
            self.main_output = MainOutput(self.settings)

        # Store paths to all PP-related files.
        self.files = {
            "acqs": "postprocessing/acquisitions.dat",
            "min_preds": "postprocessing/minimum_predictions.dat",
            "conv_meas": "postprocessing/convergence_measures.dat",
            "hypers": "postprocessing/hyperparameters.dat",
            "truehat": "postprocessing/true_f_at_x_hat.dat",
        }

        # Define a dict to hold all data structures used to generate the PP output.
        # The fundamental items are acqs, mod_par and min_preds: these must be extracted
        # from a BOMain object or read from file and can the be used to generate
        # xnexts, est_yranges and conv_meas.
        self.data = {
            "acqs": None,
            "mod_par": None,
            "min_preds": None,
            "xnexts": None,
            "est_yranges": None,
            "conv_meas": None,
        }

        # create needed directories
        if os.path.isdir("postprocessing"):
            print("warning: overwriting directory 'postprocessing'")

        shutil.rmtree("postprocessing", ignore_errors=True)
        os.makedirs("postprocessing", exist_ok=True)

        if self.settings["pp_models"]:
            os.makedirs("postprocessing/data_models", exist_ok=True)
            os.makedirs("postprocessing/graphs_models", exist_ok=True)
        if self.settings["pp_acq_funcs"]:
            os.makedirs("postprocessing/data_acqfns", exist_ok=True)
            os.makedirs("postprocessing/graphs_acqfns", exist_ok=True)
        if self.settings["pp_local_minima"] is not None:
            os.makedirs("postprocessing/data_local_minima", exist_ok=True)

    @property
    def slc_dim(self):
        sts = self.settings
        _slc_dim = 1 if sts["pp_model_slice"][0] == sts["pp_model_slice"][1] else 2
        return _slc_dim

    def run(self):
        self.load_data()
        self.dump_data()
        self.plot()

    def load_data(self):
        sts = self.settings
        dim = sts.dim
        initpts = sts["initpts"]
        iterpts = sts["iterpts"]
        minfreq = sts["minfreq"]
        totalpts = initpts + iterpts
        bo_results = self.bo_results

        # Obtain the basic output data: acqs, mod_bar, min_preds from BoMain,
        # or from boss out and rst files.
        if bo_results is not None:
            acqs = np.c_[np.arange(1, totalpts + 1), bo_results.X, bo_results.Y]

            mod_par = np.c_[np.arange(initpts, totalpts + 1), bo_results.hyperparams]

            # Determine loop indices for which convergence data was calculated.
            if 0 < minfreq < totalpts:
                imins = np.arange(initpts, totalpts, minfreq)
                imins = np.append(imins, totalpts)
            else:
                imins = np.array([totalpts])

            min_preds = np.c_[
                imins,
                np.array(list(bo_results.convergence[:, 2]), dtype=np.float64),
                bo_results.convergence[:, 3:],
            ]

        else:
            input_data = parse.parse_input_file(self.rstfile, skip="keywords")
            acqs, mod_par = split_rst_data(input_data["rst_data"], sts.dim)
            min_preds = parse.parse_min_preds(self.settings, self.outfile)

        # Generate the derived output data: xnexts, conv_meas and est_yranges.
        xnexts = []
        for i in range(len(acqs) - 1):
            xnexts.append([])
            xnexts[-1].append(acqs[i, 0])
            for n in range(dim):
                xnexts[-1].append(acqs[i + 1, 1 + n])
        xnexts = np.array(xnexts)

        est_yranges = np.array(
            [
                [acqs[i - 1, 0], max(acqs[:i, -1]) - min(acqs[:i, -1])]
                for i in range(1, len(acqs) + 1)
            ]
        )

        conv_meas = None
        if min_preds is not None:
            conv_meas = np.array(
                [
                    [
                        min_preds[i, 0],
                        euclidean(
                            min_preds[i - 1, 2 : dim + 2], min_preds[i, 2 : dim + 2],
                        ),
                        abs(min_preds[i - 1, -2] - min_preds[i, -2])
                        / est_yranges[
                            _find_index(est_yranges, int(min_preds[i, 0])), -1
                        ],
                    ]
                    for i in range(1, len(min_preds))
                ]
            )

        self.data.update(
            {
                "conv_meas": conv_meas,
                "est_yranges": est_yranges,
                "xnexts": xnexts,
                "acqs": acqs,
                "mod_par": mod_par,
                "min_preds": min_preds,
            }
        )

    def dump_data(self):
        """Dump standard data. """
        acqs = self.data["acqs"]
        mod_par = self.data["mod_par"]
        min_preds = self.data["min_preds"]
        conv_meas = self.data["conv_meas"]
        totalpts = self.settings["initpts"] + self.settings["iterpts"]

        # Initialize dump files
        # 1. acquisitions
        ioutils.overwrite(
            self.files["acqs"],
            "# Data acquisitions " + "by iteration (iter npts x y)\n",
        )
        # 2. Minimum predictions
        if min_preds is not None:
            ioutils.overwrite(
                self.files["min_preds"],
                "# Global minimum predictions by iteration"
                + " (iter npts x_hat mu_hat nu_hat)\n",
            )

        # 3. Convergence measures
        if conv_meas is not None:
            ioutils.overwrite(
                self.files["conv_meas"],
                "# Convergence measures by iteration"
                + " (iter npts dx_hat abs(dmu_hat)/yrange)\n",
            )

        # 4. Hyperparameters
        ioutils.overwrite(
            self.files["hypers"],
            "# Model hyperparameter values by iteration"
            + " (iter npts variance lengthscales)\n",
        )
        # 5. True function values at xhats
        if self.settings["pp_truef_at_xhats"] and min_preds is not None:
            ioutils.overwrite(
                self.files["truehat"],
                "# True function value at x_hat locations"
                + " by iteration (iter npts f(x_hat) "
                + "f(x_hat)-mu_hat)\n",
            )

        for npts in range(1, totalpts + 1):
            it = npts - self.settings["initpts"]

            # acquisitions
            ind_acqs = _find_index(acqs, npts)
            if ind_acqs is not None:
                ioutils.append_write(
                    self.files["acqs"],
                    ioutils.data_line([it, npts], acqs[ind_acqs, 1:], fstr="%18.10E"),
                )

            # hyperparameters
            ind_par = _find_index(mod_par, npts)
            if ind_par is not None:
                ioutils.append_write(
                    self.files["hypers"],
                    ioutils.data_line([it, npts], mod_par[ind_par, 1:], fstr="%18.10E"),
                )

            # global minimum predictions
            ind_mp = None
            if min_preds is not None:
                ind_mp = _find_index(min_preds, npts)
                if ind_mp is not None:
                    ioutils.append_write(
                        self.files["min_preds"],
                        ioutils.data_line(
                            [it, npts], min_preds[ind_mp, 1:], fstr="%18.10E"
                        ),
                    )

            # convergence measures
            if conv_meas is not None:
                ind_cm = _find_index(conv_meas, npts)
                if ind_cm is not None:
                    ioutils.append_write(
                        self.files["conv_meas"],
                        ioutils.data_line(
                            [it, npts, conv_meas[ind_cm, -2], conv_meas[ind_cm, -1],],
                            fstr="%18.10E",
                        ),
                    )

            # true function at xhat
            if (
                self.settings["pp_truef_at_xhats"]
                and min_preds is not None
                and ind_mp is not None
            ):
                self.main_output.progress_msg("Evaluating true function at x_hat", 2)
                tfhat = self.settings.f(
                    np.atleast_2d(min_preds[ind_mp, 1 : self.settings.dim + 1])
                )
                muhat = min_preds[ind_mp, -2]
                ioutils.append_write(
                    self.files["truehat"],
                    ioutils.data_line([it, npts, tfhat, tfhat - muhat], fstr="%18.10E"),
                )

    def plot(self, target="all"):
        if target in ["hyperparameters", "all"]:
            self._plot_hypers()

        # Check that minimum predictions are present in the data, otherwise
        # we cannot plot any of the targets below.
        # TODO: Should implement post-run generation of min preds.
        # if self.data['min_preds'] is None:
        #     print(f'BOSS must be run with minfreq > 0 to plot target: {target}')
        #     return

        if target in ["acquisitions", "all"]:
            self._plot_acqs()

        if target in ["convergence", "all"]:
            self._plot_conv()

        if target in ["model", "all"]:
            self._plot_model()

        if target in ["truef", "all"]:
            self._plot_truef()

    def _plot_hypers(self):
        # plot standard data dumps
        hypers = np.atleast_2d(np.loadtxt(self.files["hypers"], skiprows=1))
        if len(hypers) > 1:
            plot.plot_hyperparameters(
                self.settings, "postprocessing/hyperparameters.png", hypers
            )

    def _plot_acqs(self):
        minp = np.atleast_2d(np.loadtxt(self.files["min_preds"], skiprows=1))

        if len(minp[0]) == 0:
            raise ValueError("No minimal points found in out-file.")
        acqs_data = np.atleast_2d(np.loadtxt(self.files["acqs"], skiprows=1))

        if len(acqs_data) > 1:
            plot.plot_data_acquisitions(
                self.settings,
                "postprocessing/acquisition" + "_locations.png",
                acqs_data,
                minp,
            )

    def _plot_conv(self):
        minp = np.atleast_2d(np.loadtxt(self.files["min_preds"], skiprows=1))

        if len(minp[0]) == 0:
            raise ValueError("No minimal points found in out-file.")

        conv_meas = np.atleast_2d(np.loadtxt(self.files["conv_meas"], skiprows=1))
        if len(conv_meas) > 1:
            plot.plot_conv_measures(
                self.settings, "postprocessing/convergence_measures.png", conv_meas
            )

        if self.settings["pp_truef_at_xhats"]:
            truef_hats = np.atleast_2d(np.loadtxt(self.files["truehat"], skiprows=1))
            if len(truef_hats) > 1:
                plot.plot_truef_hat(
                    self.settings,
                    "postprocessing/true_function" + "_at_xhats.png",
                    truef_hats,
                )

    def _plot_model(self):
        acqs = self.data["acqs"]
        mod_par = self.data["mod_par"]
        conv_meas = self.data["conv_meas"]
        min_preds = self.data["min_preds"]
        xnexts = self.data["xnexts"]
        sts = self.settings
        slc_dim = self.slc_dim

        # recreate snapshots of model in a loop for all pp_iters
        curr_xhat = None

        for it in sts["pp_iters"]:

            npts = sts["initpts"] + it
            ind_acqs = _find_index(acqs, npts)
            ind_par = _find_index(mod_par, npts)

            if ind_acqs is None or ind_par is None:
                self.main_output.progress_msg(
                    "Could not find data or "
                    + " parameters to recreate model at iteration %i," % (it)
                    + " skipping post-processing for that iteration.",
                    0,
                )
                continue

            self.main_output.progress_msg("Post-processing iteration %i" % (it), 1)
            bo = recreate_bo(sts, acqs[: ind_acqs + 1, 1:], mod_par[ind_par, 1:])
            assert npts == bo.model.X.shape[0], "Model recreate fail!"

            # find current xhat
            ind_mp = _find_index(min_preds, npts)
            if ind_mp is not None:
                curr_xhat = min_preds[ind_mp, 1 : sts.dim + 1]
            ind_cm = _find_index(conv_meas, npts)

            # Local minima
            if sts["pp_local_minima"] is not None:
                self.main_output.progress_msg("Finding model local minima", 2)
                mins = Minimization.minimize(
                    bo.model.mu_with_grad,
                    sts["bounds"],
                    sts["kernel"],
                    np.hstack([bo.get_x(), bo.get_y()]),
                    sts["min_dist_acqs"],
                    accuracy=sts["pp_local_minima"],
                    args=(),
                    lowest_min_only=False,
                )

                mins = sorted(mins, key=lambda x: (x[1]))
                minima_data = [[]] * (sts.dim + 2)
                for m in mins:
                    p = []
                    for i in range(sts.dim):
                        p.append(m[0][i])
                    p.append(bo.get_mu(m[0]))
                    p.append(bo.get_nu(m[0]))
                    minima_data = np.insert(minima_data, len(minima_data[0]), p, axis=1)

                titleLine = "# Local minima (x mu nu) - model data ensemble size %i" % (
                    npts
                )
                ioutils.write_cols(
                    "postprocessing/data_local_minima/"
                    "it%.4i_npts%.4i.dat" % (it, npts),
                    minima_data,
                    titleLine=titleLine,
                )
            # Model (cross-sections)
            if sts["pp_models"]:
                dump.dump_model(
                    sts,
                    "postprocessing/data_models/it%.4i_npts%.4i.dat" % (it, npts),
                    bo,
                    bo.model.get_all_params(),
                    curr_xhat,
                )
                mdata = np.loadtxt(
                    "postprocessing/data_models/" + "it%.4i_npts%.4i.dat" % (it, npts),
                    skiprows=2,
                )
                if ind_mp is not None:
                    xhat = min_preds[ind_mp, 1 : sts.dim + 1]
                else:
                    xhat = None
                macqs = acqs[: ind_acqs + 1, 1:]
                if slc_dim < sts.dim and slc_dim == 1:
                    macqs = None
                ind_xnext = _find_index(xnexts, npts)
                if ind_xnext is not None and slc_dim == sts.dim:
                    xnext = xnexts[ind_xnext, 1:]
                else:
                    xnext = None
                if sts["pp_local_minima"] is not None and slc_dim == sts.dim:
                    minima = np.atleast_2d(np.array(minima_data)).T
                else:
                    minima = None

                plot.plot_model(
                    sts,
                    "postprocessing/graphs_models/it%.4i_npts%.4i" ".png" % (it, npts),
                    mdata,
                    xhat,
                    macqs,
                    xnext,
                    minima,
                )

            # acquisition function (cross-sections)
            if sts["pp_acq_funcs"] and it >= 0:
                ind_xnext = _find_index(xnexts, npts)
                if ind_xnext is not None:
                    defs = xnexts[ind_xnext, 1:]
                else:
                    defs = curr_xhat
                xn, acqfn = bo._acqnext(it)
                dump.dump_acqfn(
                    sts,
                    "postprocessing/data_acqfns/it%.4i_npts%.4i.dat" % (it, npts),
                    bo,
                    acqfn,
                    defs,
                )
                acqfn_data = np.loadtxt(
                    "postprocessing/data_acqfns/" + "it%.4i_npts%.4i.dat" % (it, npts),
                    skiprows=1,
                )
                if ind_mp is not None and slc_dim == sts.dim:
                    xhat = min_preds[ind_mp, 1 : sts.dim + 1]
                else:
                    xhat = None
                macqs = acqs[: ind_acqs + 1, 1:] if slc_dim == 2 else None
                ind_xnext = _find_index(xnexts, npts)
                if ind_xnext is not None:
                    xnext = xnexts[ind_xnext, 1:]
                else:
                    xnext = None

                plot.plot_acq_func(
                    sts,
                    "postprocessing/graphs_acqfns/" "it%.4i_npts%.4i.png" % (it, npts),
                    acqfn_data,
                    macqs,
                    xhat,
                    xnext,
                )

    def _plot_truef(self):
        """Dump and plot true function (cross-section). """

        sts = self.settings
        min_preds = self.data["min_preds"]
        acqs = self.data["acqs"]
        xnexts = self.data["xnexts"]
        slc_dim = self.slc_dim

        if sts["pp_truef_npts"] is None:
            # print("BOSS must be run with pp_truef_npts to plot the true function. ")
            return

        self.main_output.progress_msg("Dumping and plotting true function", 1)

        # if curr_xhat is None:
        curr_xhat = min_preds[-1, 1 : sts.dim + 1]

        dump.dump_truef(sts, "postprocessing/true_func.dat", curr_xhat)
        truef_data = np.loadtxt("postprocessing" + "/true_func.dat", skiprows=1)
        plot.plot_truef(sts, "postprocessing/true_func.png", truef_data)
        ind = np.where(min_preds[:, 1 : sts.dim + 1] == curr_xhat)[0][0]
        truef_slc_xhat_npts = int(min_preds[ind, 0])

        # replot 1D models with truef if it is now available
        if not (
            sts["pp_model_slice"][0] == sts["pp_model_slice"][1] and sts["pp_models"]
        ):
            print(
                "Replotting of 1D models with the true function requires pp_models and pp_model_slice[0]==pp_model_slice[1]"
            )
            return

        self.main_output.progress_msg("Replotting 1D models with true" + " function", 1)
        for mdat_file in os.listdir("postprocessing/data_models"):

            # find it and and npts from naming it%.4i_npts%.4i.dat
            negit = True if mdat_file[2] == "-" else False
            it = int(mdat_file[2:6]) if not negit else int(mdat_file[2:7])
            npts = int(mdat_file[-8:-4])
            mdata = np.loadtxt(
                "postprocessing/data_models/" + "it%.4i_npts%.4i.dat" % (it, npts),
                skiprows=2,
            )
            ind_mp = _find_index(min_preds, npts)
            if ind_mp is not None:
                xhat = min_preds[ind_mp, 1 : sts.dim + 1]
            else:
                xhat = None
            ind_acqs = _find_index(acqs, npts)
            macqs = acqs[: ind_acqs + 1, 1:]
            if slc_dim < sts.dim and slc_dim == 1:
                macqs = None
            ind_xnext = _find_index(xnexts, npts)
            if ind_xnext is not None and slc_dim == sts.dim:
                xnext = xnexts[ind_xnext, 1:]
            else:
                xnext = None

            if sts["pp_local_minima"] is not None and slc_dim == sts.dim:
                minima = np.loadtxt(
                    "postprocessing/data_local_"
                    + "minima/it%.4i_npts%.4i.dat" % (it, npts),
                    skiprows=1,
                )
                minima = np.atleast_2d(minima)
            else:
                minima = None

            if npts != truef_slc_xhat_npts and slc_dim < sts.dim:
                truef_d = None
            else:
                truef_d = truef_data

            plot.plot_model(
                sts,
                "postprocessing/graphs_models/" + "it%.4i_npts%.4i.png" % (it, npts),
                mdata,
                xhat,
                macqs,
                xnext,
                minima,
                truef_d,
            )



import copy
import numpy as np


class BOResults:
    """Provides easy access to the most important results from a BOSS run.

    This is a wrapper class around BOMain that provides access to the
    main results from an optimization run. In addition to defining methods for convenient
    evaluation of the predictive mean, variance and acquisition function of the GP model,
    the following optimization results are stored as attributes:

    Attributes:
    -----------
    exit_ok: bool
        Whether BOSS exited successfully, i.e. reached convergence or the maximum
        number of iterations.
    X: np.ndarray
        2D array containing all the X-data used during the optimization, this includes
        both initial random points as well as acquistions.
    Y: np.ndarray
        2D array containing the user function evaluations corresponding to the X-data.
    fmin: float
        The global minimum value as determined by BOSS.
    xmin: np.ndarray
        The global minimum, as determined by BOSS.
    xbest: The x-acquisition with the smallest user function value out
            of all the acquisitions.
    ybest: The user function value corresponding to xbest.
    fevals: The number of times the user function was evaluated.
    hyperparams: The final values of the unfixed hyperparameters in the GP model.
    """

    def __init__(self, bo_main):
        """Iinitializes the results from a BOMain object. """
        self.bo = bo_main
        bo = self.bo
        self.exit_ok = bo_main.exit_ok
        self.settings = bo.settings
        self.X = self.bo.get_x()
        self.Y = self.bo.get_y()

        self.convergence = bo.convergence
        if bo.convergence.shape[0] > 0:
            self.fmin = bo.convergence[-1, 3]
            self.xmin = bo.convergence[-1, 2]
            self.xbest = bo.convergence[-1, 0]
            self.ybest = bo.convergence[-1, 1]
        else:
            self.fmin = None
            self.xmin = None
            self.xbest = None
            self.ybest = None

        self.fevals = max(self.X.shape[0] - self.settings["initpts"], 0)
        self.hyperparams = bo.unfixed_model_params

    def f(self, x):
        """Evaluates the predictive mean of the GP model.

        Parameters
        ----------
        x: np.ndarray
            A single x-value given as a 1d array, or multiple x-values
            given as rows in a 2d array.

        Returns
        -------
        float or np.ndarray
            The mean value(s) corresponding to x. If single x-value was given this
            is a scalar and for multiple x-values an 1d array is returned.
        """
        nd = max(np.ndim(x), 1)
        return self.bo.get_mu(x).T[(0,) * (3 - nd)]

    def fvar(self, x):
        """Evaluates the predictive variance of the GP model.

        Parameters
        ----------
        x: np.ndarray
            A single x-value given as a 1d array, or multiple x-values
            given as rows in a 2d array.

        Returns
        -------
        float or np.ndarray
            The variance value(s) corresponding to x. If single x-value was given this
            is a scalar and for multiple x-values an 1d array is returned.
        """
        nd = max(np.ndim(x), 1)
        return self.bo.get_nu(x).T[(0,) * (3 - nd)]

    def acqfn(self, x):
        """Evaluates the acquisition function using the final GP model.

        Parameters
        ----------
        x: np.ndarray
            A single x-value given as a 1d array, or multiple x-values
            given as rows in a 2d array.

        Returns
        -------
        float or np.ndarray
            The acquisiton function value(s) corresponding to x. If single x-value was given this
            is a scalar and for multiple x-values an 1d array is returned.
        """
        bo = self.bo
        nd = max(np.ndim(x), 1)
        return bo.acqfn(x, bo.model, bo.acqfnpars)[0].T[(0,) * (3 - nd)]

    def copy(self):
        return copy.deepcopy(self)



import time


class Timer:
    """
    Utility class for wall clock time keeping.
    """

    def __init__(self):
        """
        Initializes the timer and starts counting time.
        """
        self.start = time.time()
        self.lapstart = self.start

    def startLap(self):
        """
        Starts counting lap time from zero.
        """
        self.lapstart = time.time()

    def getTotalTime(self):
        """
        Returns total time since timer creation in seconds.
        """
        return time.time() - self.start

    def getLapTime(self):
        """
        Returns lap time in seconds.
        """
        return time.time() - self.lapstart

    def str_totalTime(self, units="s"):
        """
        Returns total time since timer creation in the given unit of time
        as a string.
        """
        t = self.getTotalTime()
        returnString = ""
        if units == "s":
            returnString += "%12.3f" % (t)
        elif units == "m":
            returnString += "%12.3f" % (t / 60)
        elif units == "h":
            returnString += "%6.1f" % (t / 3600)
        elif units == "d":
            returnString += "%4.1f" % (t / (3600 * 24))
        return returnString

    def str_lapTime(self, units="s"):
        """
        Returns lap time in the given unit of time as a string.
        """
        t = self.getLapTime()
        returnString = ""
        if units == "s":
            returnString += "%12.3f" % (t)
        elif units == "m":
            returnString += "%12.3f" % (t / 60)
        elif units == "h":
            returnString += "%6.1f" % (t / 3600)
        elif units == "d":
            returnString += "%4.1f" % (t / (3600 * 24))
        return returnString



import numpy as np
from scipy.optimize import minimize
from scipy.stats import gamma


class Distributions:
    """
    Utilities for handling distributions.
    """

    @staticmethod
    def gammaparams(q1, q2, p1=0.1, p2=0.9):
        """ A function for parametrizing gamma distributions by specifying two
        quantiles.
         
        Relevant math can be found in "Determining distribution parameters from
        quantiles" by John D. Cook, 2010.

        :param q1: p1-quantile
        :param q2: p2-quantile
        :param p1: lower percentage (0 < p1 < 1)
        :param p2: higher percentage (0 < p2 < 1)
        """
        # Require the arguments to be positive and q1 < q2, p1 < p2 <1
        if q1 <= 0 or q2 <= q1:
            return (None, None)
        if p1 <= 0 or p2 <= p1 or p2 >= 1:
            return (None, None)

        def f(a):
            return np.abs(
                gamma.ppf(p2, a, scale=1) / gamma.ppf(p1, a, scale=1) - q2 / q1
            )

        shape = minimize(f, 1, bounds=[[0.1, None]], method="L-BFGS-B").x
        rate = gamma.ppf(p1, shape, scale=1) / q1

        return (shape[0], rate[0])



import warnings

warnings.filterwarnings("ignore")  # ignore warnings
import numpy as np
import copy
from operator import itemgetter

import numpy as np
import scipy.optimize
from scipy.spatial.distance import euclidean

import boss.utils.sobol_seq as sobol_seq


class Minimization:
    """
    Minimization utilities
    """

    @staticmethod
    def _run_minimizer(func, x0, bounds, args, finite_grads=False):
        """
        Runs a single L-BFGS-B minimization starting from x0
        """
        if finite_grads:
            x, f, d = scipy.optimize.fmin_l_bfgs_b(
                func,
                x0,
                bounds=bounds,
                args=args,
                factr=1e8,
                pgtol=1e-4,
                maxfun=1e4,
                maxiter=1e4,
                maxls=15,
                approx_grad=True,
            )
            return x, f, d["warnflag"] == 0

        else:
            x, f, d = scipy.optimize.fmin_l_bfgs_b(
                func,
                x0,
                bounds=bounds,
                args=args,
                factr=1e8,
                pgtol=1e-4,
                maxfun=1e4,
                maxiter=1e4,
                maxls=15,
            )
            return x, f, d["warnflag"] == 0

    @staticmethod
    def minimize(
        func,
        bounds,
        kerntype,
        acqs,
        min_dist_acqs,
        accuracy=0.3,
        args=(),
        lowest_min_only=True,
    ):
        """
        Tries to find global minimum of func by starting minimizers from
        accuracy*100 percentage of lowest acquisitions. func has to return both
        value and gradient given an x of same length as bounds.
        """
        acqsx = np.array(copy.deepcopy(acqs[:, : len(bounds)]))
        acqsy = np.array(copy.deepcopy(acqs[:, -1:]))

        points = []
        for i in range(len(acqsy)):
            points.append([list(acqsx[i]), float(acqsy[i])])

        points = Minimization._remove_duplicates(points, min_dist_acqs)
        points = sorted(points, key=itemgetter(1))
        X, Y = zip(*points)
        ind_last = min(len(Y) - 1, max(0, round(accuracy * (len(Y) - 1))))
        X = X[: ind_last + 1]
        X = np.array([X[i] for i in range(len(X))])

        widebounds = bounds.copy()
        p = np.array(kerntype) == "stdp"
        l = bounds[:, 1] - bounds[:, 0]
        widebounds[p, 0] -= 0.1 * l[p]
        widebounds[p, 1] += 0.1 * l[p]

        minima = []
        for acq in X:
            # Run bounded minimization on bounds which have been stretched for
            # periodic dimensions, then return the minima which have been found
            # inside the bounds. This prevents minima from being falsely
            # identified at the periodic boundaries.
            x, f, s = Minimization._run_minimizer(func, acq, widebounds, args)
            if s and np.all((bounds[p, 0] <= x[p]) & (x[p] <= bounds[p, 1])):
                minima.append([list(x), float(f)])

        if len(minima) == 0:
            lo_acq = [list(acqsx[np.argmin(acqsy)]), float(np.min(acqsy))]
            minima.append(lo_acq)

        if lowest_min_only:
            globalmin = minima[0]
            for minimum in minima:
                if minimum[1] < globalmin[1]:
                    globalmin = minimum
            return globalmin
        else:
            minima = Minimization._remove_duplicates(minima, min_dist_acqs)
            return minima

    @staticmethod
    def minimize_from_sobol(
        func, bounds, num_pts, shift, args=(), lowest_min_only=True
    ):
        """
        Tries to find global minimum of func by starting minimizers from
        num_pts shifted sobol points. func has to return both
        value and gradient given an x of same length as bounds.
        """
        sobs = np.transpose(sobol_seq.i4_sobol_generate(len(bounds), num_pts, 1))

        points = []
        for i in range(len(sobs)):
            p = []
            for n in range(len(bounds)):
                bl = bounds[n][1] - bounds[n][0]
                p.append(bounds[n][0] + (sobs[i, n] * bl + shift[n]) % bl)
            points.append(p)

        minima = []
        for pnt in points:
            x, f, s = Minimization._run_minimizer(func, pnt, bounds, args)
            if s:
                minima.append([list(x), float(f)])
        if len(minima) == 0:
            err_min = [list(points[-1]), float(func(points[-1], args[0], args[1])[0])]
            minima.append(err_min)  # later a better solution here?

        if lowest_min_only:
            globalmin = minima[0]
            for minimum in minima:
                if minimum[1] < globalmin[1]:
                    globalmin = minimum
            return globalmin
        else:
            return minima

    @staticmethod
    def minimize_from_random(
        func, bounds, num_pts, shift, args=(), lowest_min_only=True
    ):
        """
        Tries to find global minimum of func by starting minimizers from
        num_pts random points (min 10 max 100). func has to return both
        value and gradient given an x of same length as bounds.
        """
        if num_pts < 10:
            num_pts = 10
        elif num_pts > 100:
            num_pts = 100

        rands = np.random.random([num_pts, len(bounds)])

        points = []
        for i in range(len(rands)):
            p = []
            for n in range(len(bounds)):
                bl = bounds[n][1] - bounds[n][0]
                p.append(bounds[n][0] + rands[i, n] * bl)
            points.append(p)

        minima = []
        for pnt in points:
            x, f, s = Minimization._run_minimizer(func, pnt, bounds, args)
            if s:
                minima.append([list(x), float(f)])
        if len(minima) == 0:
            err_min = [list(points[-1]), float(func(points[-1], args[0], args[1])[0])]
            minima.append(err_min)  # later a better solution here?

        if lowest_min_only:
            globalmin = minima[0]
            for minimum in minima:
                if minimum[1] < globalmin[1]:
                    globalmin = minimum
            return globalmin
        else:
            return minima

    @staticmethod
    def _remove_duplicates(original, min_distance):
        """
        Removes duplicates from a list of found local minima given a minimum
        distance between distinct minima.
        """
        if len(original) == 0:
            return original
        current = copy.deepcopy(original)
        next = []
        firstRef = current[0][0]
        first_encountered_times = 0
        while True:
            next = []
            ref = current[0][0]
            ref_f = current[0][1]
            if ref == firstRef:
                first_encountered_times += 1
                if first_encountered_times == 2:
                    break
            for i in range(1, len(current)):
                if euclidean(current[i][0], ref) > min_distance:
                    next.append(current[i])
            next.append([ref, ref_f])
            current = next
        return current



"""
The format of the input file is a set of space-separated keyword value pairs:

    # comment
    key1 val1
    key2 val2  # comment2

Keyword values are evaluated according to the ruleset:

    foo               -> str
    3                 -> int
    1.0               -> float
    foo bar           -> List[str]
    3 10              -> np.ndarray(shape(1, 2), dtype=np.int64)
    5.0 7.0           -> np.ndarray(shape(1, 2), dtype=np.float64)
    1.0 2.0 ; 3.0 4.0 -> np.ndarray(shape=(2,2), dtype=np.float64)
"""
import time
import importlib
import inspect
import os
import sys
from pathlib import Path
import boss.keywords as bkw
import itertools

import numpy as np


def parse_input_file(file_path, skip=None):
    """Parses a boss input file.

    Parses the contents of the file and organizes keywords and their
    values into a dictionary. If present, restart data present under
    the optional results header is also parsed.

    Parameters
    ----------
    input_file : Optional[Union[str, Path]]
        Input file path.

    skip: Optional[str] = None
        Skip parsing part of the input, can be either 'keywords' or 'results'.

    Returns
    ----------
    input_data: dict
        A dictionary containing the parsed input data. Can contain
        the following items:

        is_rst : bool
            Whether rst data was present in the input_file or not

        rst_data: np.ndarray
            Array containing rst data, each row is composed of the X, y data
            followed by the model parameters. Rows containing the initial
            points and/or missing y-data are padded with np.nan.
    """
    input_data = {"is_rst": False}
    with open(file_path, "r") as fd:
        text = fd.read()

    # The strategy is to split the text around the results header
    # and then parse the two parts seprately.
    text_split = text.split("RESULTS:")

    # The first part of the split contains all boss keywords.
    input_data["keywords"] = {}
    if skip != "keywords":
        keywords = input_data["keywords"]
        kw_text = text_split[0].strip()
        kw_lines = kw_text.splitlines()
        for ln in kw_lines:
            ln = ln.split("#", maxsplit=1)[0].rstrip()
            if len(ln) > 0:
                key, val_str = ln.split(maxsplit=1)
                category = bkw.find_category(key)
                keywords[key] = bkw.destringify(val_str, category)

    # The second part of the split contains the restart data.
    if len(text_split) == 2:
        input_data["is_rst"] = True
        if skip != "results":
            rst_text = text_split[1].strip()
            rst_lines = rst_text.splitlines()

            # If no actual rst data is present we return early.
            if len(rst_lines) == 0:
                return input_data

            arrs = []
            for ln in rst_lines:
                ln = ln.split("#", maxsplit=1)[0].rstrip()
                if len(ln) > 0:
                    arrs.append(np.fromstring(ln, dtype=float, sep=" "))

            totalpts = len(arrs)
            max_len = max([len(a) for a in arrs])
            rst_data = np.full((totalpts, max_len), np.nan)
            for i, arr in enumerate(arrs):
                rst_data[i, : len(arr)] = arr

            input_data["rst_data"] = rst_data
    return input_data


def parse_config():
    """Parses a config file for BOSS keywords to use as new defaults.

    The name of the config file must be .bossrc or bossrc. BOSS will search
    for a config file under the following directories, in order:

    $PWD > $HOME > $HOME/.config/boss

    Only the first file encountered will be parsed.

    Returns
    ----------
    keywords: dict
        BOSS keywords parsed from the config file.
    """
    parents = [Path.cwd(), Path.home(), Path.home() / ".config/boss"]
    stems = [".bossrc", "bossrc"]
    keywords = {}
    for parent, stem in itertools.product(parents, stems):
        path = parent / stem
        if path.is_file():
            keywords = parse_input_file(path)["keywords"]
            break
    return keywords


def parse_min_preds(settings, outfile):
    """
    Extracts xhat, muhat and nuhat from all bo iterations in the output
    file. Returns a numpy array where each row is [npts, xhat, muhat,
    nuhat]. Works on output files with all verbosity levels.
    """
    data = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("Global minimum prediction") != -1:
                line = f.readline()
                line = line.split()
                data.append(np.concatenate([[npts], line[: settings.dim + 2]]))
            line = f.readline()

    return np.array(data).astype(float)


def parse_acqs(settings, outfile):
    """
    Extracts data acquisitions from all bo iterations in the output file.
    Returns a numpy array where each element is [npts, x, y].
    Note that there may be several acquisitions on the same iteration -
    especially the initial points at the 0th iteration.
    Works on output files with all verbosity levels.
    """
    acq = []
    npts = 0

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Data point added to dataset") != -1:
                line = f.readline()
                while len(line.split()) == settings.dim + 1 and line.find(".") != -1:
                    line = line.split()
                    npts += 1
                    acq.append(np.concatenate([[npts], line[: settings.dim + 1]]))
                    line = f.readline()
            else:
                line = f.readline()

    return np.array(acq).astype(float)


def parse_best_acqs(settings, outfile):
    """
    Extracts xbest and ybest from all bo iterations in the output file.
    Returns a numpy array where each element is [npts, xbest, ybest].
    Works only on output files with verbosity levels of at least 1.
    """
    data = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("Best acquisition") != -1:
                line = f.readline()
                line = line.split()
                data.append(np.concatenate([[npts], line[: settings.dim + 1]]))
            line = f.readline()

    return np.array(data).astype(float)


def parse_conv_measures(settings, outfile):
    """
    Extracts dx_hat and dmu_hat from all bo iterations in the output
    file. Returns a numpy array where each element is
    [npts, dxhat, dmuhat].
    Works only on output files with verbosity levels of at least 1.
    """
    data = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("Global minimum convergence") != -1:
                line = f.readline()
                line = line.split()
                if "none" in line:
                    pass
                else:
                    data.append(np.concatenate([[npts], line]))
            line = f.readline()

    return np.array(data).astype(float)


def parse_mod_params(outfile):
    """
    Extracts unfixed GP model hyperparameters from all bo iterations in the
    output file. Returns a numpy array where each element is
    [npts, variances, lengthscales].
    Works only on output files with verbosity levels of at least 2.
    """
    hpar = []
    npts = None

    with open(outfile, "r") as f:
        line = f.readline()
        while len(line) > 0:
            if line.find("Total ensemble size") != -1:
                npts = int(line.split()[-1])
            elif line.find("GP model hyperparameters") != -1:
                line = f.readline()
                line = line.split()
                line = list(np.insert(line, 0, line[-1]))
                del line[-1]
                hpar.append(np.concatenate([[npts], line]))
            line = f.readline()

    return np.array(hpar).astype(float)


def parse_xnexts(outfile):
    """
    Extracts xnext locations from all bo iterations in the output file.
    Returns a numpy array where each element is [npts, xnext].
    Note that this means that the location 'xnext' is to be evaluated
    in the next iteration 'iter+1'.
    Works only on output files with verbosity levels of at least 2.
    """
    try:
        data = []
        with open(outfile) as f:
            data = []
            line = f.readline()
            npts = None
            while len(line) > 0:
                if line.find("Total ensemble size") != -1:
                    npts = int(line.split()[-1])
                elif line.find("Next sampling location") != -1:
                    line = f.readline()
                    line = line.split()
                    data.append(np.concatenate([[npts], line]))
                line = f.readline()
            return np.array(data).astype(float)
    except:
        raise OSError("Error trying to read file '" + outfile + "'")


def parse_minima(outfile):
    """
    Extract local minima
    """
    try:
        f = open(outfile)
        line = f.readline()
        if line[:14] == "# Local minima":
            ok = True
        else:
            ok = False
        f.close()
    except:
        raise OSError("Error trying to read file '" + outfile + "'")

    if not ok:
        raise Exception("'" + outfile + "' not recognized as " + "a minima file.")

    try:
        f = open(outfile)
        line = f.readline()
        line = f.readline()
        minima = np.array(line.split())
        line = f.readline()
        while len(line) > 0:
            minima = np.vstack((minima, line.split()))
            line = f.readline()
        f.close()
        return minima.astype(float)

    except:
        raise OSError("Error trying to read file '" + outfile + "'")



import os
import shutil

from boss.mep.mep import MEP
from boss.mep.space import Space
import boss.io.dump as dump
import boss.io.ioutils as ioutils
import boss.io.parse as parse
from boss.pp.pp_main import recreate_bo
import boss.pp.plot as plot


class MEPMain:
    def __init__(self, settings, ipt_rstfile, minimafile, main_output):
        # create needed directories
        if os.path.isdir("mep"):
            print("warning: overwriting directory 'mep'")
        shutil.rmtree("mep", ignore_errors=True)
        os.makedirs("mep", exist_ok=True)

        # recreate model and read local minima
        self.get_model(settings, ipt_rstfile, main_output)
        self.get_minima(minimafile, main_output)
        self.get_space(settings)

        # initialize and run
        mep = MEP(
            self.bo,
            self.space,
            self.minima,
            settings["mep_precision"],
            settings["mep_rrtsteps"],
            settings["mep_nebsteps"],
            settings["mep_maxe"],
        )
        mep.run_mep(main_output)

        # write to file
        for path in mep.fullpaths:
            dump.dump_mep(path)

        # plot
        if self.minima.shape[1] == 2:
            self.plot2D(settings, mep)

    def get_model(self, settings, ipt_rstfile, main_output):
        acqs, mod_par = parse.parse_rst(settings, ipt_rstfile)
        self.bo = recreate_bo(
            settings, acqs[:, 1:], mod_par[mod_par.shape[0] - 1, 1:], main_output,
        )

    def get_minima(self, minimafile, main_output):
        self.minima = parse.parse_minima(minimafile)
        self.minima = self.minima[:, :-2]

    def get_space(self, settings):
        bounds = np.transpose(settings["bounds"])
        pbc = np.array(settings["kernel"]) == "stdp"
        if not np.all(settings["periods"] == (bounds[1, :] - bounds[0, :])):
            print("warning: MEP currently assumes periods to match " + "boundlength")
        self.space = Space(bounds, pbc)

    def plot2D(self, settings, mep):
        it = np.max(settings["pp_iters"])
        npts = settings["initpts"] + it
        fname = "postprocessing/data_models/" + "it%.4i_npts%.4i.dat" % (it, npts)
        if not files_ok([fname]):
            print(
                "Model data of the last iteration is required for "
                + "automatic 2D plotting, check\nthe 'pp_models' "
                + "and 'pp_iters' options, then try rerunning postprocessing."
            )
            return
        mdata = ioutils.read_cols(fname, skiprows=2)
        xhat = None
        xnext = None
        minima = self.minima
        truef = None

        plot.model(
            settings,
            "mep/minpaths.png",
            mdata,
            minima=self.minima,
            incl_uncert=False,
            paths=mep.fullpaths,
        )


def files_ok(filenames):
    """
    Checks that the given files exist and can be opened.
    """
    for fname in filenames:
        try:
            f = open(fname, "r")
            f.close()
        except FileNotFoundError:
            print("Could not find file '" + fname + "'")
            return False
    return True



import numpy as np
from boss.mep.forest import Forest
from boss.mep.neb import neb
from boss.mep.node import Node
from boss.mep.path import Path


class MEP:
    def __init__(
        self,
        bo,
        space,
        min_points,
        precision=25,
        rrtsteps=10000,
        nebsteps=10,
        maxE=None,
    ):
        """
        Parameters:
        bo          The surrogate model class.
        space       (Space) The Space class defines boundaries and periodicity.
        min_points  (2d numpy array) Local minima of the surrogate model.
        precision   (float, optional) How many points would be in the returned
                    path, if it went from start to end of the shortest edge of
                    the space. Assumed to be 100.
        rrtsteps    (int, optional) Maximum number of RRT steps, assumed to be
                    1000. More equals better resolution for the algorithm and
                    longer runtime.
        nebsteps    (int, optional) Maximum number of NEB steps, assumed to be
                    1000. More doesn't necessarily improve performance, as the
                    NEB optimization should converge before this is reached.
        maxE        (float, optional) Global maximum energy of the model. If
                    not specified defaults to the highest observed energy.
        """
        self.bo = bo
        self.space = space
        self.precision = precision
        self.rrtsteps = rrtsteps
        self.nebsteps = nebsteps
        self.eps = (
            1 / precision * np.min(self.space.bounds[1, :] - self.space.bounds[0, :])
        )
        self.min_points = self.space.crds(np.atleast_2d(min_points))
        if min_points.shape[0] < 2:
            raise Exception("At least two minima needed to find paths.")

        e = self.bo.get_mu(self.min_points)[:, 0]
        e.sort()
        self.e_start = e[1]  # Second lowest local minimum
        if maxE == None:
            maxE = np.max(self.bo.get_y())

        self.maxE = maxE
        e_range = maxE - e[1]
        self.stepsize = e_range / rrtsteps

    def run_mep(self, main_output):
        main_output.mep_start(self)

        main_output.progress_msg("Running RRT:s", 1)
        self.get_connections()

        main_output.progress_msg(
            "Running NEB for " + str(len(self.connections)) + " paths.", 1
        )
        self.get_paths()

        main_output.mep_result(self)

    def get_connections(self):
        max_e = self.e_start
        connections = []
        forests = []
        for i in range(self.min_points.shape[0]):
            forests.append(Forest([Node(self.min_points[i, :], None, i)]))

        while len(forests) > 1:
            i = 0
            while i < len(forests):
                forests[i].grow(self.bo, self.eps, max_e, self.space)
                j = i + 1
                while j < len(forests):
                    end_nodes = forests[i].test_connectivity_once(
                        forests[j], self.bo, self.space, self.eps, max_e
                    )
                    if end_nodes != None:
                        connections.append(end_nodes)
                        forests[i].merge(forests[j])
                        forests.remove(forests[j])
                    else:
                        j += 1
                i += 1
            max_e += self.stepsize

        self.connections = connections

    def get_paths(self):
        # extend and optimize with NEB
        paths = []
        I = 0
        for connection in self.connections:
            I += 1
            path = Path(self.space)
            path.initFromConnection(connection)
            ns = int(self.nebsteps / 2)
            for i in range(2):
                path.reform(self.eps)
                path.crds = neb(path.crds, self.bo, self.space, ns)
            paths.append(path)
        self.paths = np.array(paths)

        # get paths from each minima to each other minima
        self.fullpaths = []
        for i in range(self.min_points.shape[0]):
            for j in range(self.min_points.shape[0]):
                if i < j:
                    crds = self.get_fullcrds(i, j)
                    energy = self.bo.get_mu(crds)
                    path = Path(self.space, crds, i, j, energy)
                    self.fullpaths.append(path)

    def get_fullcrds(self, i, j, notchecked=None):
        if i == j:
            return None
        if np.any(notchecked == None):
            notchecked = np.array(np.ones(self.paths.shape[0], dtype=bool))

        has_i = [path.mi == i or path.mj == i for path in self.paths]
        check = np.logical_and(has_i, notchecked)
        checkind = np.array(range(self.paths.shape[0]))[check]
        notchecked = np.logical_and(notchecked, np.logical_not(check))
        if checkind.shape[0] == 0:
            return None

        found = [path.mi == j or path.mj == j for path in self.paths[checkind]]
        if np.any(found):
            path = self.paths[checkind[found]][0]
            if path.mj == j:
                return path.crds
            else:
                return np.flipud(path.crds)
        else:
            for ind in checkind:
                if self.paths[ind].mi == i:
                    newi = self.paths[ind].mj
                else:
                    newi = self.paths[ind].mi
                fullcrds = self.get_fullcrds(newi, j, notchecked)
                if np.all(fullcrds != None):
                    path = self.paths[ind]
                    if self.paths[ind].mi == i:
                        return np.vstack((path.crds, fullcrds))
                    else:
                        return np.vstack((np.flipud(path.crds), fullcrds))



import GPy


class KernelFactory:
    """
    This class contains the construction of the kernel.
    """

    @staticmethod
    def construct_kernel(settings, forced_hypers=None):
        """
        Creates the kernel.
        """
        kerns = [None] * (settings.dim)
        KernelFactory._select_kernels(kerns, settings, forced_hypers)
        if forced_hypers is None:
            KernelFactory._set_constraints(kerns, settings)
        if forced_hypers is None:
            KernelFactory._set_priors(kerns, settings)

        # multiplies the kernels into one object and returns it
        Kernel = kerns[0]
        if len(kerns) > 1:
            for i in range(1, len(kerns)):
                Kernel = Kernel * kerns[i]
        return Kernel

    @staticmethod
    def _select_kernels(kerns, settings, forced_hypers):
        """
        Selects and creates kernel objects for each dimension. Hyperparameters
        are set to their initial values and default constraints removed.
        """
        for i in range(settings.dim):
            if forced_hypers is None:
                if i == 0:
                    ksi = settings["thetainit"][0]
                else:
                    ksi = 1.0
                klsi = settings["thetainit"][i + 1]
            else:
                ksi = forced_hypers[0]
                klsi = forced_hypers[1 + i]
            kper = settings["periods"][i]

            ktype = settings["kernel"][i]
            if ktype == "stdp":
                kerns[i] = GPy.kern.StdPeriodic(
                    input_dim=1,
                    variance=ksi,
                    period=kper,
                    lengthscale=klsi,
                    ARD1=True,
                    ARD2=True,
                    active_dims=[i],
                    name="kern",
                )
            elif ktype == "rbf":
                kerns[i] = GPy.kern.RBF(
                    input_dim=1,
                    variance=ksi,
                    lengthscale=klsi,
                    ARD=True,
                    active_dims=[i],
                    name="kern",
                )
            elif ktype == "mat32":
                kerns[i] = GPy.kern.Matern32(
                    input_dim=1,
                    variance=ksi,
                    lengthscale=klsi,
                    ARD=True,
                    active_dims=[i],
                    name="kern",
                )
            elif ktype == "mat52":
                kerns[i] = GPy.kern.Matern52(
                    input_dim=1,
                    variance=ksi,
                    lengthscale=klsi,
                    ARD=True,
                    active_dims=[i],
                    name="kern",
                )
            else:
                raise TypeError(f"""Unknown kernel {settings['kernel'][i]}""")

    #            kerns[i].unconstrain()

    @staticmethod
    def _set_constraints(kerns, settings):
        """
        Sets hyperparameter constraints on kernels.
        """
        # variance
        if settings["thetabounds"] is not None:
            kerns[0].variance.constrain_bounded(
                settings["thetabounds"][0][0],
                settings["thetabounds"][0][1],
                warning=False,
            )
            # lengthscale
            for i in range(settings.dim):
                kerns[i].lengthscale.constrain_bounded(
                    settings["thetabounds"][i + 1][0],
                    settings["thetabounds"][i + 1][1],
                    warning=False,
                )
        # period
        for i in range(settings.dim):
            if settings["kernel"][i] == "stdp":  # pbc
                kerns[i].period.constrain_fixed(settings["periods"][i], warning=False)

        # other than the first kernel's variances
        if settings.dim > 1:
            for i in range(1, settings.dim):
                kerns[i].variance.constrain_fixed(1.0, warning=False)

    @staticmethod
    def _set_priors(kerns, settings):
        """
        Sets hyperparameter priors on kernels.
        """
        if settings["thetaprior"] is not None:
            prior = None
            if settings["thetaprior"] == "gamma":
                prior = GPy.priors.Gamma
            else:
                raise TypeError(
                    "Unknown value '"
                    + settings["thetaprior"]
                    + "' given in keyword thetaprior."
                )

            # variance
            kerns[0].variance.set_prior(
                prior(settings["thetapriorpar"][0][0], settings["thetapriorpar"][0][1]),
                warning=False,
            )
            # lengthscale
            for i in range(settings.dim):
                kerns[i].lengthscale.set_prior(
                    prior(
                        settings["thetapriorpar"][i + 1][0],
                        settings["thetapriorpar"][i + 1][1],
                    ),
                    warning=False,
                )



import numpy as np
from boss.mep.node import Node


class Forest:
    def __init__(self, nodes):
        self.trees = nodes

    def get_closest_point(self, point, space):
        closest = self.trees[0]
        smallest_dist = np.linalg.norm(space.vec(closest.coords, point))
        for tree in self.trees:
            temp = tree.get_closest_point(point, space)
            if temp[1] < smallest_dist:
                closest = temp[0]
                smallest_dist = temp[1]
        return (closest, smallest_dist)

    def get_all_children(self):
        all_children = []
        for tree in self.trees:
            all_children.extend(tree.get_all_children())
        return all_children

    def grow(self, bo, eps, max_e, space):
        boundlength = space.bounds[1, :] - space.bounds[0, :]
        dim = bo.dim
        new_point = np.random.random(dim) * boundlength + space.bounds[0, :]
        new_e = bo.get_mu(np.vstack(np.atleast_2d(new_point)))[0, 0]
        if new_e < max_e:
            closest, dist = self.get_closest_point(new_point, space)
            new_point = space.trace(closest.coords, new_point, bo, eps, max_e)
            if not np.prod(new_point == closest.coords):
                closest.children.append(Node(new_point, closest))

    def get_random_child(self):
        all_children = self.get_all_children()
        return np.random.choice(all_children, size=1)[0]

    def merge(self, other_forest):
        for tree in other_forest.trees:
            self.trees.append(tree)

    def test_connectivity_once(self, forest_B, bo, space, eps, max_e):
        random_A = self.get_random_child()
        closest_B, dist = forest_B.get_closest_point(random_A.coords, space)
        endpoint = space.trace(random_A.coords, closest_B.coords, bo, eps, max_e)
        if np.prod(endpoint == closest_B.coords):
            return [random_A, closest_B]
        else:
            return None



import re
from os import path

from setuptools import find_packages, setup


def parse_metadata(
    file_path, extra_keywords=None, add_dunder=True, find_author_list=False
):
    """Parse core Python package metadata stored in a file. """

    metadata = {}
    if extra_keywords:
        metadata.update({k: None for k in extra_keywords})
    metadata.update(
        {
            "version": None,
            "url": None,
            "author": None,
            "author_email": None,
            "maintainer": None,
            "maintainer_email": None,
            "license": None,
            "description": None,
        }
    )

    with open(file_path) as fd:
        text = fd.read()

    # The following regex will match the contents of any non-multiline string.
    # Note: Surrounding quotes are NOT included.
    re_str_tmpl = "{}\s*=\s*['\"](.*)['\"]"

    for key in metadata.keys():
        if add_dunder:
            var_name = "__{}__".format(key)
        else:
            var_name = key
        match = re.search(re_str_tmpl.format(var_name), text)
        if match:
            metadata[key] = match.group(1)

    # Extra: If no author was found, look for a python list named authors or _authors__
    # and use it to build the author string.
    if find_author_list and not metadata["author"]:
        match = re.search("__authors__\s*=\s*[\[\(](.*)[\)\]]", text, re.DOTALL)
        if match:
            stringified_list = match.group(1)
            author_list = stringified_list.split(",")
            author_list = [auth.strip()[1:-1] for auth in author_list]
            metadata["author"] = ", ".join(author_list)

    metadata = {k: v for k, v in metadata.items() if v is not None}
    return metadata


metadata = parse_metadata("boss/__init__.py", find_author_list=True)

with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Physics",
    "Topic :: Scientific/Engineering :: Mathematics",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "License :: OSI Approved :: Apache Software License",
]

setup(
    name="aalto-boss",
    version=metadata["version"],
    url=metadata["url"],
    author=metadata["author"],
    maintainer=metadata["maintainer"],
    maintainer_email=metadata["maintainer_email"],
    description=metadata["description"],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    license=metadata["license"],
    include_package_data=True,
    classifiers=classifiers,
    packages=find_packages(),
    install_requires=["GPy", "matplotlib>=3.0"],
    python_requires=">=3.5",
    entry_points={"console_scripts": ["boss=boss.__main__:main",],},
)



import time
import importlib
import inspect
import os
import sys
from pathlib import Path
import boss.keywords as bkw

import numpy as np


def twoDfloatarray_line(arr, shape1, shape2, formatString=" %8.3E"):
    if arr is None:
        return " none"
    arr = np.array(arr)
    s = ""
    for i in range(shape1):
        for j in range(shape2):
            s += formatString % (arr[i, j])
        s += ";   "
    return s[:-4]


def oneDarray_line(arr, length, type_, float_format=" %8.3E"):
    if arr is None:
        return " none"
    if type_ is str or type_ is bool:
        formatString = " %s"
    elif type_ is int:
        formatString = " %i"
    elif type_ is float:
        formatString = float_format
    s = ""
    for i in range(length):
        s += formatString % (arr[i])
    return s


def data_line(X, Q=None, fstr="%15.7E", separ="     "):
    """
    Returns a line of data. X and Q (optional) should be
    1D arrays/lists.
    """
    s = ""
    X = np.atleast_1d(X)
    for i in range(len(X)):
        if X[i] is not None:
            s += fstr % (X[i])
        else:
            fmt = "%" + str(_format_space(fstr)) + "s"
            s += fmt % ("none")
    if Q is not None:
        s += separ
        Q = np.atleast_1d(Q)
        for i in range(len(Q)):
            if Q[i] is not None:
                s += fstr % (Q[i])
            else:
                fmt = "%" + str(_format_space(fstr)) + "s"
                s += fmt % ("none")
    return s + "\n"


def append_write(file_path, text):
    """
    Writes the text into the given file appending.
    """
    f = open(file_path, "a")
    f.write(text)
    f.close()


def overwrite(file_path, text):
    """
    Writes the text into the given file overwriting.
    """
    f = open(file_path, "w")
    f.write(text)
    f.close()


def write_cols(
    file_path, vectorList, space="    ", titleLine=None, formatString="%15.7E"
):
    """
    Needs as input a vector of form [ [column1 elements], [column2 elements], ...  ]
    """
    f = open(file_path, "w")

    if titleLine is not None:
        f.write(titleLine + "\n")

    num_cols = len(vectorList)
    num_elements = len(
        vectorList[0]
    )  # number of elements must be the same in all columns

    for elem in range(num_elements):
        for col in range(num_cols):
            item = vectorList[col][elem]
            if item == "none":
                fmt = formatString[:-3] + "s"
                f.write(fmt % (item) + space)
            else:
                f.write(formatString % (float(item)) + space)

        f.write("\n")

    f.close()


def read_cols(file_path, skiprows=1):
    """
    Reads data columns from a file into a numpy array, where each column is
    separated like col1 = arr[:,0], col2 = arr[:,1] etc.
    """
    return np.loadtxt(file_path, skiprows=skiprows)


def _format_space(fmt_str, error_default=15):
    fmt_str = fmt_str.replace(" ", "")
    if fmt_str[0] == "%" and len(fmt_str) > 2:
        fmt_str = fmt_str[1:-1]
        return int(fmt_str.split(".")[0])
    else:
        return error_default



import GPy
import numpy as np

from boss.utils.minimization import Minimization


class Model(GPy.core.gp.GP):
    """
    Functionality for creating, refitting and optimizing a GP model
    """

    def __init__(
        self, X, Y, bounds, min_dist_acqs, minzacc, kerntype, kernel, dim, noise
    ):
        """
        Initializes the Model class.
        """
        self.bounds = bounds
        self.min_dist_acqs = min_dist_acqs
        self.minzacc = minzacc
        self.kerntype = kerntype
        self.kernel = kernel
        self.dim = dim
        super().__init__(
            X,
            Y,
            self.kernel,
            likelihood=GPy.likelihoods.Gaussian(),
            inference_method=GPy.inference.latent_function_inference.ExactGaussianInference(),
        )
        self.likelihood.fix(noise)

    def add_data(self, newX, newY):
        """
        Updates the model evidence (observations) dataset appending.
        """
        X = np.vstack([self.X, np.atleast_2d(newX)])
        Y = np.vstack([self.Y, newY])
        self.set_XY(X, Y)

    def redefine_data(self, X, Y):
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        self.set_XY(np.atleast_2d(X), np.atleast_2d(Y))

    def optimize_controlrstr(self, restarts=1):
        """
        Updates the model hyperparameters by maximizing marginal likelihood.
        """
        self.optimization_runs = []
        if restarts == 1:
            super().optimize()
        else:
            super().optimize_restarts(
                num_restarts=restarts, verbose=False, messages=False
            )

    def get_unfixed_params(self):
        """
        Returns the unfixed parameters of the model in an array where
        the first element is the variance and the rest are lenghtscales.
        Note that this function assumes that only variances and lengthscales
        are free hyperparameters.
        """
        return np.array(self.unfixed_param_array.copy()).astype(float)

    def get_all_params(self):
        """
        Returns the parameters of the model in format:
        noise, variance, lengthscales, periods
        where the last two are 1D lists. There exists a period only for those
        dimensions which are using a periodic kernel.
        """
        pars = self.param_array
        noise = float(pars[-1])
        sigma = float(pars[0])
        ls_and_per = pars[1:-1]
        lss = []
        pers = []
        pbc_dims = list(np.where(np.array(self.kerntype) == "stdp")[0])
        ind = 0
        for i in range(self.dim):
            if i in pbc_dims:
                pers.append(float(ls_and_per[ind]))
                ind += 1
            lss.append(float(ls_and_per[ind]))
            ind += 2  # jump over fixed var
        return noise, sigma, lss, pers

    def set_unfixed_params(self, params):
        """
        Sets the unfixed parameters of the model (variance, lengthscales) to
        given values.
        """
        self[".*kern.variance"] = params[0]
        lss = self[".*lengthscale*"]
        for i in range(self.dim):
            lss[i] = params[i + 1]
        self.trigger_update()

    def min_prediction(self):
        """
        Find and returns the model global minimum prediction xhat
        and the model mean (muhat) and variance (nuhat) at that point.
        """
        gmin = Minimization.minimize(
            self.mu_with_grad,
            self.bounds,
            self.kerntype,
            np.hstack([self.X, self.Y]),
            self.min_dist_acqs,
            accuracy=self.minzacc,
        )
        xhat = np.array(gmin[0])
        muhat, nuhat = self.predict(np.atleast_2d(xhat))
        return xhat, muhat[0][0], nuhat[0][0]

    def mu(self, x):
        """
        Returns model prediction mean at point x.
        """
        m, v = self.predict(np.atleast_2d(x))
        return m

    def nu(self, x):
        """
        Returns model prediction variance at point x, taking into account the
        model variance (noise).
        """
        m, v = self.predict(np.atleast_2d(x))
        v = np.clip(v, 1e-12, np.inf)
        return np.sqrt(v)

    def mu_with_grad(self, x):
        """
        Returns model mean and its gradient at point x.
        """
        m, v = self.predict(np.atleast_2d(x))
        dmdx, dvdx = self.predictive_gradients(np.atleast_2d(x))
        scipygradient = np.asmatrix(dmdx).transpose()
        return m, scipygradient

    def get_bestxy(self):
        """
        Returns the lowest energy acquisition (x, y). Note that this does not
        account for the model mean shift introduced by the BOMain class.
        """
        xbest = np.array(self.X[np.argmin(self.Y)])
        ybest = np.min(self.Y)
        return xbest, ybest

    def predictive_m_s_grad(self, x):
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, taking into account the model variance (noise).
        """
        m, v = self.predict(np.atleast_2d(x))
        v = np.clip(v, 1e-12, np.inf)
        dmdx, dvdx = self.predictive_gradients(np.atleast_2d(x))
        dmdx = dmdx[:, :, 0]
        dsdx = dvdx / (2 * np.sqrt(v))
        return m, np.sqrt(v), dmdx, dsdx



"""
Plotting functionality. All plot functions make PNG images to destination
file based on given source array(s) of data.
"""

import numpy as np
import matplotlib.pyplot as plt

plt.switch_backend("agg")
import warnings


def plot_data_acquisitions(settings, dest_file, acqs, min_preds):
    """
    Plots minimum predictions and uncertainty, acquisition energies 
    and locations as a function of iteration.
    """
    itNo_acqs = np.clip(acqs[:, 0], int(min_preds[0, 0]), np.inf)
    x = acqs[:, 2 : settings.dim + 2]
    y = acqs[:, settings.dim + 2]
    itNo_mp = min_preds[:, 0]
    xhat = min_preds[:, 2 : settings.dim + 2]
    muhat = min_preds[:, -2]
    nuhat = min_preds[:, -1]

    plt.subplot(211)
    plt.fill_between(itNo_mp, muhat + nuhat, muhat, color="lightgrey")
    plt.fill_between(itNo_mp, muhat - nuhat, muhat, color="lightgrey")
    plt.plot(itNo_mp, muhat + nuhat, "grey", linewidth=2)
    plt.plot(itNo_mp, muhat - nuhat, "grey", linewidth=2)
    plt.plot(itNo_mp, muhat, "black", linewidth=5, label=r"$\mu(\hat{x})$")
    plt.scatter(
        itNo_acqs,
        y,
        s=200,
        linewidth=4,
        facecolors="none",
        edgecolors="black",
        label="$acq_y$",
    )

    plt.xlim(min(itNo_mp), max(itNo_mp))
    plt.xticks(itNo_mp[:: max(1, round(len(itNo_mp) / 10))])
    plt.ylabel("$y$ with $\mu(\hat{x})$", size=20)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)

    plt.subplot(212)
    colors = [
        "blue",
        "red",
        "green",
        "brown",
        "yellow",
        "purple",
        "cyan",
        "magenta",
        "grey",
        "gold",
    ]
    for i in range(x.shape[1]):
        plt.plot(
            itNo_mp,
            xhat[:, i],
            color=colors[i % len(colors)],
            linewidth=5,
            label="$\hat{x}_%i$" % (i + 1),
        )
        plt.scatter(
            itNo_acqs,
            x[:, i],
            s=200,
            linewidth=4,
            facecolors="none",
            edgecolors=colors[i % len(colors)],
            label="$acq_{x%i}$" % (i + 1),
        )

    plt.xlim(min(itNo_mp), max(itNo_mp))
    plt.xticks(itNo_mp[:: max(1, round(len(itNo_mp) / 10))])
    plt.xlabel("iteration", size=20)
    plt.ylabel("$x_i$ with $\hat{x}_i$", size=20)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)

    plt.suptitle("Data acquisitions", size=20)
    plt.gcf().set_size_inches(9, 8)
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.savefig(dest_file)
    plt.close()


def plot_conv_measures(settings, dest_file, conv_meas, legends=True):
    """
    Plots quantities related to convergence as a function of iteration.
    """
    itNo = conv_meas[:, 0]
    dxhat = conv_meas[:, 2]
    dmuhat = conv_meas[:, 3]

    plt.subplot(211)
    plt.plot(itNo, dxhat, "k", linewidth=5)
    # plt.xlabel('iteration', size=20)
    plt.ylabel(r"$\Delta \hat{x}$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(dxhat) > 0:
        plt.yscale("log")

    plt.subplot(212)
    plt.plot(itNo, dmuhat, "k", linewidth=5)
    plt.xlabel("iteration", size=20)
    plt.ylabel(r"$abs(\Delta \mu(\hat{x}))/\Delta y$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(dmuhat) > 0:
        plt.yscale("log")

    plt.suptitle("Convergence tracking", size=20)
    plt.gcf().set_size_inches(9, 7)
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.savefig(dest_file)
    plt.close()


def plot_hyperparameters(settings, dest_file, hypers, legends=True):
    """
    Plots GP model unfixed hyperparameters (variances and lengthscales)
    as a function of iteration.
    """
    itNo = hypers[:, 0]
    var = hypers[:, 2]
    ls = hypers[:, 3:]

    colors = [
        "blue",
        "red",
        "green",
        "brown",
        "yellow",
        "purple",
        "cyan",
        "magenta",
        "black",
        "gold",
    ]
    markers = ["s", "p", "*", "h", "^", "+", "x", "d", "v", "|"]

    plt.subplot(211)
    plt.plot(itNo, var, color="r", linewidth=5)
    plt.ylabel("Variance", size=20)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.yscale("log")
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    # plt.gca().ticklabel_format(useOffset=False)

    plt.subplot(212)
    for i in range(settings.dim):
        plt.plot(
            itNo,
            ls[:, i],
            color=colors[i % 10],
            linewidth=5,
            linestyle="-",
            marker=markers[i % 10],
            markersize=12,
            label=r"$\ell_ %i$" % (i + 1),
        )
    plt.xlabel("iteration", size=20)
    plt.ylabel("Lengthscale", size=20)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.yscale("log")
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    # plt.gca().ticklabel_format(useOffset=False)
    if legends:
        lgd = plt.legend(loc="upper right", prop={"size": 18})

    plt.suptitle("Hyperparameter values", size=20)
    plt.gcf().set_size_inches(11, 7)
    # plt.tight_layout(rect=[0, 0.03, 1, 0.95]); plt.savefig(dest_file)
    if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
        plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
    else:
        plt.savefig(dest_file)
    plt.close()


def plot_truef_hat(settings, dest_file, truef_hats, legends=True):
    """
    Plots true function value at xhat locations
    as a function of iteration.
    """
    itNo = truef_hats[:, 0]
    truehat = truef_hats[:, -2]
    last = truehat[-1]
    best = np.array([np.min(truehat[:i]) for i in range(1, len(truehat) + 1)])
    tfhat_muhat = truef_hats[:, -1]

    plt.subplot(211)
    plt.plot(itNo, abs(truehat - last), "k", linewidth=5)
    plt.plot(itNo, abs(best - last), "k", linestyle="dashed", linewidth=3, label="best")
    plt.grid(True)
    plt.ylabel(r"$f(\hat{x})-f(\hat{x}_{last})$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(abs(truehat - last)) and max(abs(best - last)) > 0:
        plt.yscale("log")
    if legends:
        lgd = plt.legend(loc="upper right", prop={"size": 20})

    plt.subplot(212)
    plt.plot(itNo, abs(tfhat_muhat), "k", linewidth=5)
    plt.xlabel("iteration", size=20)
    plt.ylabel(r"$f(\hat{x})-\mu(\hat{x})$", size=24)
    plt.xticks(itNo[:: max(1, round(len(itNo) / 10))])
    plt.grid(True)
    plt.gca().tick_params(labelsize=18)
    plt.gca().ticklabel_format(useOffset=False)
    if max(abs(tfhat_muhat)) > 0:
        plt.yscale("log")

    plt.suptitle("True function at minimum predictions", size=20)
    plt.gcf().set_size_inches(9, 9)
    # plt.tight_layout(rect=[0, 0.03, 1, 0.95]); plt.savefig(dest_file)
    if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
        plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
    else:
        plt.savefig(dest_file)
    plt.close()


def plot_model(
    settings,
    dest_file,
    model_data,
    xhat=None,
    acqs=None,
    xnext=None,
    minima=None,
    truef=None,
    incl_uncert=True,
    axis_labels=None,
    legends=True,
    paths=None,
):
    """
    Plots a (max 2D) slice of the model.
    """
    coords = model_data[:, : settings.dim]
    mu, nu = model_data[:, -2], model_data[:, -1]
    slice_dim = (
        1 if settings["pp_model_slice"][0] == settings["pp_model_slice"][1] else 2
    )
    if axis_labels is None:
        axis_labels = [
            "$x_%i$" % (settings["pp_model_slice"][0] + 1),
            "$x_%i$" % (settings["pp_model_slice"][1] + 1),
        ]

    if slice_dim == 1:
        x1 = coords[:, settings["pp_model_slice"][0]]
        if truef is not None:
            crds = truef[:, : settings.dim]
            tf = truef[:, -1]
            tfx1 = crds[:, settings["pp_model_slice"][0]]
            plt.plot(tfx1, tf, "k", linewidth=5, label="f(x)")
        if incl_uncert:
            plt.fill_between(x1, mu + nu, mu, color="lightgrey")
            plt.fill_between(x1, mu - nu, mu, color="lightgrey")
            plt.plot(x1, mu + nu, "grey", linewidth=3, label="$\\nu(x)$")
            plt.plot(x1, mu - nu, "grey", linewidth=3)
        plt.plot(x1, mu, "b", linewidth=5, label="$\\mu(x)$")

        if xhat is not None:
            plt.axvline(
                xhat[settings["pp_model_slice"][0]],
                color="red",
                linewidth=5,
                label="$\hat{x}$",
                zorder=19,
            )

        if acqs is not None:
            x1 = acqs[:, settings["pp_model_slice"][0]]
            y = acqs[:, -1]
            plt.scatter(
                x1,
                y,
                s=200,
                linewidth=6,
                facecolors="none",
                edgecolors="brown",
                label="acqs",
                zorder=18,
            )

        if xnext is not None:
            plt.axvline(
                xnext[settings["pp_model_slice"][0]],
                color="green",
                linewidth=5,
                label="$x_{next}$",
                linestyle="dashed",
                zorder=20,
            )

        if minima is not None:
            x1 = minima[:, settings["pp_model_slice"][0]]
            y = minima[:, -2]
            plt.scatter(
                x1,
                y,
                s=250,
                linewidth=6,
                zorder=19,
                facecolors="none",
                edgecolors="lawngreen",
                label="minima",
            )

        plt.xlim(
            min(coords[:, settings["pp_model_slice"][0]]),
            max(coords[:, settings["pp_model_slice"][0]]),
        )
        yd = max(mu) - min(mu)
        plt.ylim(min(mu) - 0.1 * yd, max(mu) + 0.1 * yd)
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel("$y$", size=24)
        if legends:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        # plt.tight_layout()
        if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()

    elif slice_dim == 2:
        npts = settings["pp_model_slice"][2]
        x1 = coords[:, settings["pp_model_slice"][0]]
        x2 = coords[:, settings["pp_model_slice"][1]]

        if truef is not None:
            pass  # 2D true func slice will be plotted in separate graph only

        if incl_uncert:
            plt.contour(x1[:npts], x2[::npts], nu.reshape(npts, npts), 25, colors="k")
            plt.contourf(
                x1[:npts], x2[::npts], nu.reshape(npts, npts), 150, cmap="viridis"
            )
            cbar = plt.colorbar()  # , orientation='horizontal')
            cbar.set_label(label="$\\nu(x)$", size=24)
            cbar.ax.tick_params(labelsize=18)
            plt.xlabel(axis_labels[0], size=24)
            plt.ylabel(axis_labels[1], size=24)
            plt.gcf().set_size_inches(10, 8)
            plt.gca().tick_params(labelsize=18)
            plt.tight_layout()
            plt.savefig(dest_file[:-4] + "_uncert.png")
            plt.close()

        plt.contour(x1[:npts], x2[::npts], mu.reshape(npts, npts), 25, colors="k")
        plt.contourf(x1[:npts], x2[::npts], mu.reshape(npts, npts), 150, cmap="viridis")
        cbar = plt.colorbar()  # , orientation='horizontal')
        cbar.set_label(label="$\mu(x)$", size=24)
        cbar.ax.tick_params(labelsize=18)

        lo = False
        if xhat is not None:
            plt.plot(
                xhat[settings["pp_model_slice"][0]],
                xhat[settings["pp_model_slice"][1]],
                "r*",
                markersize=26,
                zorder=21,
                label="$\hat{x}$",
            )
            lo = True

        if acqs is not None:
            x1 = acqs[:, settings["pp_model_slice"][0]]
            x2 = acqs[:, settings["pp_model_slice"][1]]
            sz = np.linspace(200, 500, len(x1))
            lw = np.linspace(3, 8, len(x1))
            plt.scatter(
                x1[0],
                x2[0],
                s=sz[int(len(x1) / 2.0)],
                linewidth=lw[int(len(x1) / 2.0)],
                zorder=10,
                facecolors="none",
                edgecolors="magenta",
                label="acqs",
            )
            for i in range(len(x1)):
                plt.scatter(
                    x1[i],
                    x2[i],
                    s=sz[i],
                    linewidth=lw[i],
                    zorder=10,
                    facecolors="none",
                    edgecolors="magenta",
                )
            lo = True

        if xnext is not None:
            plt.plot(
                xnext[settings["pp_model_slice"][0]],
                xnext[settings["pp_model_slice"][1]],
                "b^",
                markersize=26,
                label="$x_{next}$",
                zorder=20,
            )
            lo = True

        if minima is not None:
            x1 = minima[:, settings["pp_model_slice"][0]]
            x2 = minima[:, settings["pp_model_slice"][1]]
            plt.scatter(
                x1,
                x2,
                s=350,
                linewidth=6,
                facecolors="none",
                edgecolors="navajowhite",
                zorder=11,
                label="minima",
            )
            lo = True

        if paths is not None:
            threshold = min(settings["bounds"][:, 1] - settings["bounds"][:, 0]) / 2
            for path in paths:
                start = 0
                stop = 1
                while True:
                    if (
                        stop == path.crds.shape[0] - 1
                        or np.linalg.norm(path.crds[stop, :] - path.crds[stop - 1, :])
                        > threshold
                    ):
                        plt.plot(
                            path.crds[range(start, stop), 0],
                            path.crds[range(start, stop), 1],
                            linewidth=3.0,
                            color="red",
                        )
                        if stop == path.crds.shape[0] - 1:
                            break
                        else:
                            start = stop
                    stop += 1

        plt.xlim(
            min(coords[:, settings["pp_model_slice"][0]]),
            max(coords[:, settings["pp_model_slice"][0]]),
        )
        plt.ylim(
            min(coords[:, settings["pp_model_slice"][1]]),
            max(coords[:, settings["pp_model_slice"][1]]),
        )
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel(axis_labels[1], size=24)
        top = 0.99
        if legends and lo:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
            top = 0.85
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        # plt.tight_layout()
        if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()
    else:
        raise TypeError("ERROR: Model plot only applicable up to 2D (slices)")


def plot_acq_func(
    settings,
    dest_file,
    acqf_data,
    acqs=None,
    xhat=None,
    xnext=None,
    axis_labels=None,
    legends=True,
):
    """
    Plots a (max 2D) slice of the acquisition function. Can include also
    acquisitions, xhat and xnext if those are given.
    """
    coords = acqf_data[:, : settings.dim]
    af = acqf_data[:, -1]
    slice_dim = (
        1 if settings["pp_model_slice"][0] == settings["pp_model_slice"][1] else 2
    )
    if axis_labels is None:
        axis_labels = [
            "$x_%i$" % (settings["pp_model_slice"][0] + 1),
            "$x_%i$" % (settings["pp_model_slice"][1] + 1),
        ]

    if slice_dim == 1:
        x1 = coords[:, settings["pp_model_slice"][0]]
        plt.plot(x1, af, "k", linewidth=5)

        if acqs is not None:
            x1 = acqs[:, settings["pp_model_slice"][0]]
            y = acqs[:, -1]
            plt.scatter(
                x1,
                y,
                s=200,
                linewidth=6,
                facecolors="none",
                edgecolors="brown",
                label="acqs",
            )

        if xhat is not None:
            plt.axvline(
                xhat[settings["pp_model_slice"][0]],
                color="red",
                linewidth=5,
                label="$\hat{x}$",
            )

        if xnext is not None:
            plt.axvline(
                xnext[settings["pp_model_slice"][0]],
                color="green",
                linewidth=5,
                label="$x_{next}$",
                linestyle="dashed",
            )

        plt.xlim(
            min(coords[:, settings["pp_model_slice"][0]]),
            max(coords[:, settings["pp_model_slice"][0]]),
        )
        yd = max(af) - min(af)
        plt.ylim(min(af) - 0.1 * yd, max(af) + 0.1 * yd)
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel("$acqfn(x)$", size=24)
        top = 0.99
        if legends:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
            top = 0.9
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        # plt.tight_layout()
        if legends:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()

    elif slice_dim == 2:
        npts = settings["pp_model_slice"][2]
        x1 = coords[:, settings["pp_model_slice"][0]]
        x2 = coords[:, settings["pp_model_slice"][1]]

        plt.contour(x1[:npts], x2[::npts], af.reshape(npts, npts), 25, colors="k")
        plt.contourf(x1[:npts], x2[::npts], af.reshape(npts, npts), 150, cmap="viridis")
        cbar = plt.colorbar()  # , orientation='horizontal')
        cbar.set_label(label="$af(x)$", size=24)
        cbar.ax.tick_params(labelsize=18)

        lo = False
        if acqs is not None:
            x1 = acqs[:, settings["pp_model_slice"][0]]
            x2 = acqs[:, settings["pp_model_slice"][1]]
            sz = np.linspace(200, 500, len(x1))
            lw = np.linspace(3, 8, len(x1))
            plt.scatter(
                x1[0],
                x2[0],
                s=sz[int(len(x1) / 2)],
                linewidth=lw[int(len(x1) / 2)],
                zorder=10,
                facecolors="none",
                edgecolors="magenta",
                label="acqs",
            )
            for i in range(1, len(x1)):
                plt.scatter(
                    x1[i],
                    x2[i],
                    s=sz[i],
                    linewidth=lw[i],
                    zorder=10,
                    facecolors="none",
                    edgecolors="magenta",
                )
            lo = True

        if xhat is not None:
            plt.plot(
                xhat[settings["pp_model_slice"][0]],
                xhat[settings["pp_model_slice"][1]],
                "r*",
                zorder=19,
                markersize=26,
                label="$\hat{x}$",
            )
            lo = True

        if xnext is not None:
            plt.plot(
                xnext[settings["pp_model_slice"][0]],
                xnext[settings["pp_model_slice"][1]],
                "b^",
                zorder=20,
                markersize=26,
                label="$x_{next}$",
            )
            lo = True

        plt.xlim(
            min(coords[:, settings["pp_model_slice"][0]]),
            max(coords[:, settings["pp_model_slice"][0]]),
        )
        plt.ylim(
            min(coords[:, settings["pp_model_slice"][1]]),
            max(coords[:, settings["pp_model_slice"][1]]),
        )
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel(axis_labels[1], size=24)
        top = 0.99
        if legends and lo:
            lgd = plt.legend(
                bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
                loc=3,
                ncol=4,
                mode="expand",
                borderaxespad=0.0,
                prop={"size": 20},
            )
            top = 0.9
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        # plt.tight_layout()
        if legends and lo:  # FOR OLD MATPLOTLIB COMPATIBILITY
            plt.savefig(dest_file, bbox_extra_artists=(lgd,), bbox_inches="tight")
        else:
            plt.savefig(dest_file)
        plt.close()
    else:
        raise TypeError(
            "ERROR: Acquisition function plot only" + " applicable up to 2D (slices)"
        )


def plot_truef(settings, dest_file, truef_data, axis_labels=None):
    """
    Plots a (max 2D) slice of the true function.
    """
    coords = truef_data[:, : settings.dim]
    tf = truef_data[:, -1]
    slice_dim = (
        1 if settings["pp_model_slice"][0] == settings["pp_model_slice"][1] else 2
    )
    if axis_labels is None:
        axis_labels = [
            "$x_%i$" % (settings["pp_model_slice"][0] + 1),
            "$x_%i$" % (settings["pp_model_slice"][1] + 1),
        ]

    if slice_dim == 1:
        x1 = coords[:, settings["pp_model_slice"][0]]
        plt.plot(x1, tf, "k", linewidth=5)
        plt.xlim(min(x1), max(x1))
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel("$f(x)$", size=24)
        plt.title("True function", size=20)
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        # plt.tight_layout()
        plt.tight_layout()
        plt.savefig(dest_file)
        plt.close()

    elif slice_dim == 2:
        npts = settings["pp_truef_npts"]
        x1 = coords[:, settings["pp_model_slice"][0]]
        x2 = coords[:, settings["pp_model_slice"][1]]

        plt.contour(x1[:npts], x2[::npts], tf.reshape(npts, npts), 25, colors="k")
        plt.contourf(x1[:npts], x2[::npts], tf.reshape(npts, npts), 150, cmap="viridis")
        cbar = plt.colorbar()  # , orientation='horizontal')
        cbar.set_label(label="$f(x)$", size=24)
        cbar.ax.tick_params(labelsize=18)

        plt.xlim(min(x1), max(x1))
        plt.ylim(min(x2), max(x2))
        plt.xlabel(axis_labels[0], size=24)
        plt.ylabel(axis_labels[1], size=24)
        plt.gcf().set_size_inches(10, 8)
        plt.gca().tick_params(labelsize=18)
        plt.gca().ticklabel_format(useOffset=False)
        plt.title("True function", size=20)
        # plt.tight_layout()
        plt.tight_layout()
        plt.savefig(dest_file)
        plt.close()
    else:
        raise TypeError(
            "ERROR: True function plot only" + " applicable up to 2D (slices)"
        )



import numpy as np


class Node:
    def __init__(self, coords, parent, name=None):
        self.coords = coords
        self.children = []
        self.parent = parent
        if name != None:
            self.name = name
        else:
            self.name = parent.name

    def get_closest_point(self, point, space):
        closest = self
        smallest_dist = np.linalg.norm(space.vec(self.coords, point))
        for child in self.children:
            close, dist = child.get_closest_point(point, space)
            if dist < smallest_dist:
                closest = close
                smallest_dist = dist
        return (closest, smallest_dist)

    def get_all_children(self):
        all_children = [self]
        for child in self.children:
            all_children.extend(child.get_all_children())
        return all_children

    def get_all_parents(self):
        if self.parent == None:
            return []
        else:
            parents = [self.parent]
            parents.extend(self.parent.get_all_parents())
            return parents

    def get_random_child(self):
        all_children = self.get_all_children()
        return np.random.choice(all_children, size=1)[0]
